"""Deeposlandia central command-line interface

Available choices::
    - deepo datagen [args]
    - deepo train [args]
    - deepo infer [args]
    - deepo postprocess [args]
"""

import argparse
import os

from deeposlandia import datagen, train, inference, postprocess
from deeposlandia import AVAILABLE_MODELS
from deeposlandia.datasets import AVAILABLE_DATASETS, GEOGRAPHIC_DATASETS


def datagen_parser(subparser, reference_func):
    """Add arguments focused on data generation process

    Parameters
    ----------
    subparser : argparser.parser.SubParsersAction
    reference_func : function
    """
    parser = subparser.add_parser(
        "datagen",
        help="Preprocess new datasets for automated image analysis"
    )
    add_dataset_args(parser)
    add_nb_image_args(parser)
    parser.add_argument(
        "-s",
        "--image-size",
        default=256,
        type=int,
        help=("Desired size of images (width = height)"),
    )
    parser.add_argument(
        "-T",
        "--nb-testing-image",
        type=int,
        default=0,
        help=("Number of testing images"),
    )
    parser.set_defaults(func=reference_func)


def train_parser(subparser, reference_func):
    """Add instance-specific arguments from the command line

    Parameters
    ----------
    subparser : argparser.parser.SubParsersAction
    reference_func : function
    """
    parser = subparser.add_parser(
        "train",
        help="Train convolutional neural networks for automated image analysis"
    )
    add_dataset_args(parser)
    add_nb_image_args(parser)
    add_training_args(parser)
    parser.add_argument(
        "-s",
        "--image-size",
        default=256,
        type=int,
        help=("Desired size of images (width = height)"),
    )
    parser.add_argument(
        "-e",
        "--nb-epochs",
        type=int,
        default=0,
        help=(
            "Number of training epochs (one epoch means "
            "scanning each training image once)"
        ),
    )
    parser.set_defaults(func=reference_func)


def inference_parser(subparser, reference_func):
    """Add instance-specific arguments from the command line

    Parameters
    ----------
    subparser : argparser.parser.SubParsersAction
    reference_func : function
    """
    parser = subparser.add_parser(
        "infer",
        help="Predict image labelling with a neural network model"
    )
    add_dataset_args(parser)
    add_training_args(parser)
    parser.add_argument(
        "-i",
        "--image-paths",
        required=True,
        nargs="+",
        help="Path of the image on the file system",
    )
    parser.set_defaults(func=reference_func)


def postprocess_parser(subparser, reference_func):
    """Add instance-specific arguments from the command line

    Build a high-resolution image labelling by predicting semantic segmentation
    labels on image patches, and by postprocessing resulting arrays so as to
    get geographic entities.

    Parameters
    ----------
    subparser : argparser.parser.SubParsersAction
    reference_func : function
    """
    parser = subparser.add_parser(
        "postprocess",
        help=(
            "Postprocess the neural network prediction "
            + "(for geographic datasets only)"
        )
    )
    add_dataset_args(parser, available_datasets=GEOGRAPHIC_DATASETS)
    parser.add_argument(
        "-b",
        "--batch-size",
        default=2,
        type=int,
        help=("Number of images in each inference batch"),
    )
    parser.add_argument(
        "-g",
        "--draw-grid",
        action="store_true",
        help="If specified, draw the grid that materializes the predicted tiles"
    )
    parser.add_argument(
        "-i",
        "--image-basename",
        required=True,
        help="Basename of the image within the dataset",
    )
    parser.add_argument(
        "-s",
        "--image-size",
        required=True,
        type=int,
        help="Image patch size, in pixels",
    )
    parser.set_defaults(func=reference_func)


def add_dataset_args(parser, available_datasets=AVAILABLE_DATASETS):
    """Add arguments to "parser" that are related to the dataset material
    identification on the file system

    There are:
      - "datapath" that indicates where the datasets are stored on the file
    system
      - "dataset" gives the name of the dataset (which is also the dataset
    folder name on the file system)

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser that must be "augmented"
    available_datasets : tuple
        List of available datasets for the related command
    """
    parser.add_argument(
        "-P",
        "--datapath",
        default="./data",
        help="Relative path towards data directory",
    )
    parser.add_argument(
        "-D",
        "--dataset",
        required=True,
        choices=available_datasets,
        help="Dataset type",
    )


def add_nb_image_args(parser):
    """Add arguments to "parser" that are related to the amount of preprocessed
    data to prepare

    There are:
      - "nb_training_image" indicates the number of training image to process
      - "nb_validation_image" indicates the number of images to scan at each
    validation step

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser that must be "augmented"
    available_datasets : tuple
        List of available datasets for the related command
    """
    parser.add_argument(
        "-t",
        "--nb-training-image",
        type=int,
        default=0,
        help=("Number of training images"),
    )
    parser.add_argument(
        "-v",
        "--nb-validation-image",
        type=int,
        default=0,
        help=("Number of validation images"),
    )
    parser.add_argument(
        "--nb-tiles-per-image",
        type=int,
        help=(
            "Number of tile per raw image (mandatory for geographic datasets "
            "when training and/or validation images are involved)."
        ),
    )


def add_training_args(parser):
    """Add arguments to "parser" that are related to the convolutional neural
    network model training

    Such arguments are defined as follows:
      - "model" gives the problem that is solved, either semantic segmentation
    or feature detection
      - "name" refers to the instance name, for identification on the file
    system
      - "network" is the first hyperparameter, that describes the neural
    network architecture
      - "batch_size" is the second hyperparameter, namely the number of image
    in each training/validation batch
      - "dropout" indicates the rates of neurons that must be shutted down
    during each iteration, in order to reduce overfitting (third hyperparameter)
      - "learning_rate" gives the learning rate of the training process (fourth
    hyperparameter)
      - "learning_rate_decay" gives the decay rate of the learning rate
    throughout the training process (fifth hyperparameter)

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser that must be "augmented"
    """
    parser.add_argument(
        "-M",
        "--model",
        choices=AVAILABLE_MODELS,
        required=True,
        help=(
            "Type of model to train, either "
            "'featdet' (feature detection) or 'semseg' (semantic segmentation)"
        ),
    )
    parser.add_argument(
        "-n",
        "--name",
        default="cnn",
        help=(
            "Model name that will be used for results, "
            "checkout and graph storage on file system"
        ),
    )
    parser.add_argument(
        "-N",
        "--network",
        choices=["simple", "vgg", "resnet", "inception", "unet", "dilated"],
        default=["simple"],
        nargs="+",
        help=(
            "Neural network architecture, either 'simple', 'vgg', "
            "'inception' or 'resnet' for feature detection, and 'either', "
            "'unet' or 'dilated' for semantic segmentation "
            "('simple' refers to 3 conv/pool blocks and 1 fully-connected "
            "layer; the others refer to state-of-the-art networks)"
        ),
    )
    parser.add_argument(
        "-b",
        "--batch-size",
        type=int,
        nargs="+",
        default=[50],
        help=(
            "Number of images that must be contained " "into a single batch"
        ),
    )
    parser.add_argument(
        "-d",
        "--dropout",
        nargs="+",
        default=[1.0],
        type=float,
        help="Percentage of kept neurons during training",
    )
    parser.add_argument(
        "-L",
        "--learning-rate",
        nargs="+",
        default=[0.001],
        type=float,
        help=("Starting learning rate"),
    )
    parser.add_argument(
        "-l",
        "--learning-rate-decay",
        nargs="+",
        default=[0.0001],
        type=float,
        help=("Learning rate decay"),
    )


def main():
    """Main method of the module
    """
    parser = argparse.ArgumentParser(
        prog="deepo",
        description="Deeposlandia framework for automated image analysis",
    )
    sub_parsers = parser.add_subparsers(dest="command")
    datagen_parser(sub_parsers, reference_func=datagen.main)
    train_parser(sub_parsers, reference_func=train.main)
    inference_parser(sub_parsers, reference_func=inference.main)
    postprocess_parser(sub_parsers, reference_func=postprocess.main)
    args = parser.parse_args()

    if args.func:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""Tanzania building dataset class

To add building instances into the dataset, use the load_buildings() method by
passing as the first parameter the folder that contains building informations
(images, features, items)

"""

from multiprocessing import Pool
import os

import cv2
import daiquiri
import geopandas as gpd
import numpy as np
from osgeo import gdal
from PIL import Image

from deeposlandia.datasets import GeoreferencedDataset
from deeposlandia import geometries, utils


logger = daiquiri.getLogger(__name__)

# Save png tiles without auxiliary information on disk
os.environ["GDAL_PAM_ENABLED"] = "NO"


class TanzaniaDataset(GeoreferencedDataset):
    """Tanzania building dataset, as released during the Open AI Tanzania
    challenge

    See:
    https://blog.werobotics.org/2018/08/06/welcome-to-the-open-ai-tanzania-challenge/
    The dataset is composed of 20 high-resolution images (~6-8cm/pixel, until
    76k*76k pixels), 13 being associated with geo-referenced labels, 7 for
    testing purpose.

    Attributes
    ----------
    img_size : int
        Size of the tiles into which each raw images is decomposed during
    dataset population (height=width)

    """

    BACKGROUND_COLOR = [0, 0, 0]
    COMPLETE_COLOR = [50, 200, 50]
    INCOMPLETE_COLOR = [200, 200, 50]
    FOUNDATION_COLOR = [200, 50, 50]

    def __init__(self, img_size):
        """Class constructor ; instanciates a TanzaniaDataset as a standard Dataset
        which is completed by a glossary file that describes the dataset labels
        and images

        """
        super().__init__(img_size)
        self.add_label(
            label_id=0,
            label_name="background",
            color=self.BACKGROUND_COLOR,
            is_evaluate=True,
        )
        self.add_label(
            label_id=1,
            label_name="complete",
            color=self.COMPLETE_COLOR,
            is_evaluate=True,
        )
        self.add_label(
            label_id=2,
            label_name="incomplete",
            color=self.INCOMPLETE_COLOR,
            is_evaluate=True,
        )
        self.add_label(
            label_id=3,
            label_name="foundation",
            color=self.FOUNDATION_COLOR,
            is_evaluate=True,
        )


    def _preprocess_for_training(self, image_filename, output_dir, nb_images):
        """Resize/crop then save the training & label images

        Parameters
        ----------
        image_filename : str
            Full path towards the image on the disk
        output_dir : str
            Output path where preprocessed image must be saved

        Returns
        -------
        dict
            Key/values with the filenames and label ids
        """
        raster = gdal.Open(image_filename)
        raw_img_width = raster.RasterXSize
        raw_img_height = raster.RasterYSize
        image_data = raster.ReadAsArray()
        image_data = np.swapaxes(image_data, 0, 2)
        result_dicts = []
        logger.info(
            "Image filename: %s, size: (%s, %s)",
            image_filename.split("/")[-1], raw_img_width, raw_img_height
        )

        label_filename = image_filename.replace("images", "labels").replace(
            ".tif", ".geojson"
        )
        labels = gpd.read_file(label_filename)
        labels = labels.loc[~labels.geometry.isna(), ["condition", "geometry"]]
        none_mask = [lc is None for lc in labels.condition]
        labels.loc[none_mask, "condition"] = "Complete"

        nb_attempts = 0
        image_counter = 0
        empty_image_counter = 0
        while image_counter < nb_images and nb_attempts < 2 * nb_images:
            # randomly pick an image
            x = np.random.randint(0, raw_img_width - self.image_size)
            y = np.random.randint(0, raw_img_height - self.image_size)

            tile_data = image_data[
                x:(x + self.image_size), y:(y + self.image_size)
            ]
            tile_image = Image.fromarray(tile_data)
            raster_features = geometries.get_image_features(raster)
            tile_items = geometries.extract_tile_items(
                raster_features, labels, x, y, self.image_size, self.image_size
            )
            mask = self.load_mask(tile_items, raster_features, x, y)
            label_dict = utils.build_labels(
                mask, range(self.get_nb_labels()), "tanzania"
            )
            labelled_image = utils.build_image_from_config(mask, self.labels)
            if len(tile_items) > 0:
                tiled_results = self._serialize(
                    tile_image,
                    labelled_image,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "nw",
                )
                if tiled_results:
                    result_dicts.append(tiled_results)
                image_counter += 1
                tile_image_ne = tile_image.transpose(Image.FLIP_LEFT_RIGHT)
                labelled_image_ne = labelled_image.transpose(
                    Image.FLIP_LEFT_RIGHT
                )
                tiled_results_ne = self._serialize(
                    tile_image_ne,
                    labelled_image_ne,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "ne",
                )
                if tiled_results_ne:
                    result_dicts.append(tiled_results_ne)
                image_counter += 1
                tile_image_sw = tile_image.transpose(Image.FLIP_TOP_BOTTOM)
                labelled_image_sw = labelled_image.transpose(
                    Image.FLIP_TOP_BOTTOM
                )
                tiled_results_sw = self._serialize(
                    tile_image_sw,
                    labelled_image_sw,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "sw",
                )
                if tiled_results_sw:
                    result_dicts.append(tiled_results_sw)
                image_counter += 1
                tile_image_se = tile_image_sw.transpose(Image.FLIP_LEFT_RIGHT)
                labelled_image_se = labelled_image_sw.transpose(
                    Image.FLIP_LEFT_RIGHT
                )
                tiled_results_se = self._serialize(
                    tile_image_se,
                    labelled_image_se,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "se",
                )
                if tiled_results_se:
                    result_dicts.append(tiled_results_se)
                image_counter += 1
                del tile_image_se, tile_image_sw, tile_image_ne
                del labelled_image_se, labelled_image_sw, labelled_image_ne
            else:
                if empty_image_counter < 0.1 * nb_images:
                    tiled_results = self._serialize(
                        tile_image,
                        labelled_image,
                        label_dict,
                        image_filename,
                        output_dir,
                        x,
                        y,
                        "nw",
                    )
                    if tiled_results:
                        result_dicts.append(tiled_results)
                    image_counter += 1
                    empty_image_counter += 1
            nb_attempts += 1
        del raster
        logger.info(
            "Generate %s images after %s attempts.", image_counter, nb_attempts
        )
        return result_dicts



"""Flask web application for deeposlandia
"""

import daiquiri
from flask import (
    abort,
    Flask,
    jsonify,
    redirect,
    render_template,
    request,
    send_from_directory,
)
import json
import numpy as np
import os
from PIL import Image
from werkzeug.utils import secure_filename

from deeposlandia import config, utils
from deeposlandia import AVAILABLE_MODELS
from deeposlandia.datasets import AVAILABLE_DATASETS
from deeposlandia.inference import predict


logger = daiquiri.getLogger(__name__)


PROJECT_FOLDER = config.get("folder", "project_folder")
UPLOAD_FOLDER = os.path.join(PROJECT_FOLDER, "uploads/")
PREDICT_FOLDER = os.path.join(PROJECT_FOLDER, "predicted/")
os.makedirs(UPLOAD_FOLDER, exist_ok=True)
os.makedirs(PREDICT_FOLDER, exist_ok=True)

if config.get("status", "status") == "dev":
    app = Flask(__name__)
elif config.get("status", "status") == "prod":
    app = Flask(__name__, static_folder=PROJECT_FOLDER)
else:
    raise ValueError("No defined status, please consider 'dev' or 'prod'.")
app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER
app.config["ERROR_404_HELP"] = False

ALLOWED_EXTENSIONS = set(["png", "jpg", "jpeg", "tif"])


def check_model(model):
    """Check if `model` is valid

    Parameters
    ----------
    model : str
        String to verify
    """
    if model not in AVAILABLE_MODELS:
        abort(404, "Model {} not found".format(model))


def check_dataset(dataset):
    """Check if `dataset` is valid

    Parameters
    ----------
    dataset : str
        String to verify
    """
    if dataset not in AVAILABLE_DATASETS:
        abort(404, "Dataset {} not found".format(dataset))


def allowed_file(filename):
    """Check if `filename` is really an image file name on the file system,
    *i.e.* composed of at least one '.' character, and which ends with an
    allowed extensions (namely, `jpg`, `jpeg` or `png`)

    Parameters
    ----------
    filename : str
        String to verify
    """
    return (
        "." in filename
        and filename.rsplit(".", 1)[1].lower() in ALLOWED_EXTENSIONS
    )


def recover_image_info(dataset, filename):
    """Recover image full name on the application as well as corresponding
    ground-truth labels

    Parameters
    ----------
    dataset : str
    filename : str

    Returns
    -------
    dict
        Dictionary that contains image full names (raw images and labelled
    version) and label infos

    """
    image_file = os.path.join(dataset, "images", filename + ".png")
    label_file = image_file.replace("images", "labels")
    if dataset == "mapillary":
        image_file = image_file.replace(".png", ".jpg")
    server_label_filename = os.path.join(app.static_folder, label_file)
    server_label_image = np.array(Image.open(server_label_filename))
    if dataset == "mapillary":
        image_size = "400"
    elif dataset == "aerial":
        image_size = "240"
    elif dataset == "tanzania":
        image_size = "512"
    elif dataset == "shapes":
        image_size = "64"
    else:
        raise ValueError(
            (
                "Unknown dataset. Please choose 'mapillary', "
                "'aerial', 'tanzania' or 'shapes'."
            )
        )
    with open(
        os.path.join(
            "data",
            dataset,
            "preprocessed",
            image_size,
            "validation.json",
        )
    ) as fobj:
        config = json.load(fobj)
    if not dataset == "aerial":
        actual_labels = np.unique(
            server_label_image.reshape([-1, 3]), axis=0
        ).tolist()
    else:
        actual_labels = np.unique(server_label_image).tolist()
    printed_labels = [
        (item["category"], utils.GetHTMLColor(item["color"]))
        for item in config["labels"]
        if item["color"] in actual_labels
    ]
    return {
        "image_file": image_file,
        "label_file": label_file,
        "labels": printed_labels,
    }


@app.route("/")
def index():
    """Route to application home page

    Returns
    -------
    Jinja template
        Template for application home page
    """
    return render_template("index.html")


@app.route("/predictor_demo/<string:model>/<string:dataset>")
def demo_homepage(model, dataset):
    """Route to a demo page dedicated to `model` and ̀dataset`

    Parameters
    ----------
    model : str
        Considered research problem
    dataset : str
        Considered dataset

    Returns
    -------
    Jinja template
        Template for demo web page fed with the specified model and an image
    filename
    """
    check_model(model)
    check_dataset(dataset)
    return render_template(
        dataset + "_demo.html",
        model=model,
        image_filename=os.path.join("sample_image", "raw_image.png"),
        label_filename=os.path.join("sample_image", "ground_truth.png"),
        ground_truth="",
        predicted_filename=os.path.join("sample_image", "prediction.png"),
        result="",
    )


@app.route("/predictor_demo/<string:model>/<string:dataset>/<string:image>")
def predictor_demo(model, dataset, image):
    """Route to a jsonified version of deep learning model predictions, for
    demo case

    Parameters
    ----------
    model : str
        Considered research problem
    dataset : str
        Considered dataset
    image : str
        Name of the demo image onto the server

    Returns
    -------
    dict
        Deep learning model prediction for demo page (depends on the chosen
    model, either feature detection or semantic segmentation)
    """
    logger.info("file: %s, dataset: %s, model: %s", image, dataset, model)
    image_info = recover_image_info(dataset, image)
    predictions = predict(
        [os.path.join(app.static_folder, image_info["image_file"])],
        dataset,
        model,
        output_dir=PREDICT_FOLDER,
    )
    if model == "featdet":
        predicted_image = "sample_image/prediction.png"
        predicted_labels = predictions[
            os.path.join(app.static_folder, image_info["image_file"])
        ]
    elif model == "semseg":
        predicted_image = os.path.join(
            "predicted", image_info["label_file"].split("/")[-1]
        )
        predicted_labels = predictions["labels"]
    else:
        raise ValueError(
            (
                "Unknown model, please choose amongst %s.", AVAILABLE_MODELS
            )
        )
    return render_template(
        dataset + "_demo.html",
        model=model,
        image_filename=image_info["image_file"],
        label_filename=image_info["label_file"],
        ground_truth_labels=image_info["labels"],
        predicted_filename=predicted_image,
        predicted_labels=predicted_labels,
    )


@app.route("/prediction")
def prediction():
    """Route to a jsonified version of deep learning model predictions, for
    client tool

    Returns
    -------
    dict
        Deep learning model predictions
    """
    filename = os.path.basename(request.args.get("img"))
    filename = os.path.join(app.config["UPLOAD_FOLDER"], filename)
    dataset = request.args.get("dataset")
    model = request.args.get("model")
    logger.info("file: %s, dataset: %s, model: %s", filename, dataset, model)
    predictions = predict(
        [filename],
        "mapillary",
        "semseg",
        output_dir=PREDICT_FOLDER,
    )
    return jsonify(predictions)


@app.route("/uploads/<filename>")
def send_image(filename):
    """Route to uploaded-by-client images

    Returns
    -------
    file
        Image file on the server (see Flask documentation)
    """
    return send_from_directory(app.config["UPLOAD_FOLDER"], filename)


@app.route("/load_predictor")
def load_predictor():
    """
    """
    filename = os.path.join("sample_image", "ajaccio.png")
    return render_template("predictor.html", example_image=filename)


@app.route("/predictor", methods=["POST"])
def upload_image():
    """Route to deep learning predictor that takes as an input a
    uploaded-by-client image (which is saved on the server); if the uploaded
    file is not valid, the method does a simple redirection

    Returns
    -------
    Jinja template
        Template for predictor web page fed with the uploaded image

    """
    # check if the post request has the file part
    if "file" not in request.files:
        logger.info("No file part")
        return redirect(request.url)
    fobj = request.files["file"]
    # if user does not select file, browser also
    # submit a empty part without filename
    if fobj.filename == "":
        logger.info("No selected file")
        return redirect(request.url)
    if fobj and allowed_file(fobj.filename):
        message_info = ""
        filename = secure_filename(fobj.filename)
        full_filename = os.path.join(app.config["UPLOAD_FOLDER"], filename)
        fobj.save(full_filename)
        target_size = 400
        image = Image.open(full_filename)
        image = image.resize((target_size, target_size))
        if image.mode == "RGBA":
            image = Image.merge(mode="RGB", bands=image.split()[:3])
            message_info = (
                "Warning: the uploaded image has an alpha channel "
                "(transparency feature). "
                "Here we use only RGB channels."
                )
            logger.warning(message_info)
        image.save(full_filename)
        return render_template(
            "predictor.html",
            image_name=filename,
            predicted_filename="sample_image/prediction.png",
            message=message_info
        )


@app.route("/demo_image_selector")
def demo_image_selector():
    """Route to a random server file with corresponding label information

    Returns
    -------
    dict
        Dictionary that contains four items: `image_name` is the short name of
    an image that will be displayed by the app, `image_file` is the relative
    path of the image on the server, `label_file` is the relative path of its
    labelled version on the server, and `labels` is a dictionary that
    summarizes the label information for displyaing purpose
    """
    dataset = request.args.get("dataset")
    server_folder = os.path.join(app.static_folder, dataset, "images")
    filename = np.random.choice(os.listdir(server_folder))
    return jsonify(image_name=filename.split(".")[0])



"""Metrics for measuring machine learning algorithm performances
"""

from keras import backend


def iou(actual, predicted):
    """Compute Intersection over Union statistic (i.e. Jaccard Index)

    See https://en.wikipedia.org/wiki/Jaccard_index

    Parameters
    ----------
    actual : list
        Ground-truth labels
    predicted : list
        Predicted labels

    Returns
    -------
    float
        Intersection over Union value
    """
    actual = backend.flatten(actual)
    predicted = backend.flatten(predicted)
    intersection = backend.sum(actual * predicted)
    union = backend.sum(actual) + backend.sum(predicted) - intersection
    return 1.0 * intersection / union


def iou_loss(actual, predicted):
    """Loss function based on the Intersection over Union (IoU) statistic

    IoU is comprised between 0 and 1, as a consequence the function is set as
    `f(.)=1-IoU(.)`: the loss has to be minimized, and is comprised between 0
    and 1 too

    Parameters
    ----------
    actual : list
        Ground-truth labels
    predicted : list
        Predicted labels

    Returns
    -------
    float
        Intersection-over-Union-based loss
    """
    return 1.0 - iou(actual, predicted)


def dice_coef(actual, predicted, eps=1e-3):
    """Dice coef

    See https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient
    Examples at:
      -
    https://github.com/jocicmarko/ultrasound-nerve-segmentation/blob/master/train.py#L23
      -
    https://github.com/ZFTurbo/ZF_UNET_224_Pretrained_Model/blob/master/zf_unet_224_model.py#L36


    Parameters
    ----------
    actual : list
        Ground-truth labels
    predicted : list
        Predicted labels
    eps : float
        Epsilon value to add numerical stability

    Returns
    -------
    float
        Dice coef value
    """
    y_true_f = backend.flatten(actual)
    y_pred_f = backend.flatten(predicted)
    intersection = backend.sum(y_true_f * y_pred_f)
    return (2.0 * intersection + eps) / (
        backend.sum(y_true_f) + backend.sum(y_pred_f) + eps
    )


def dice_coef_loss(actual, predicted):
    """
    Parameters
    ----------
    actual : list
        Ground-truth labels
    predicted : list
        Predicted labels

    Returns
    -------
    float
        Dice-coef-based loss
    """
    return -dice_coef(actual, predicted)



"""Launch the webapp in debug mode.

WARNING: only for development purpose!!
"""

import os

import daiquiri

from deeposlandia import utils
from deeposlandia.webapp.main import app, config


logger = daiquiri.getLogger(__name__)


if "symlink" not in config.sections():
    raise KeyError("Config.ini file does not contain any 'symlink' section.")
for link_name, path in config.items("symlink"):
    utils.create_symlink(os.path.join(app.static_folder, link_name), path)

app.config["DEBUG"] = True
app.config["TEMPLATES_AUTO_RELOAD"] = True
app.run(host="0.0.0.0", port=7897, debug=True)



""" Utilitary function for Mapillary dataset analysis
"""

import json
import os

import daiquiri
import numpy as np
import pandas as pd
from PIL import Image

from deeposlandia.datasets import GEOGRAPHIC_DATASETS


logger = daiquiri.getLogger(__name__)


def read_config(filename):
    """Read the JSON configuration file.

    Parameters
    ----------
    filename : str
        Path of the configuration file

    Returns
    -------
    dict
        Dataset glossary
    """
    with open(filename) as fobj:
        return json.load(fobj)


def build_labels(filtered_image, label_ids, dataset="mapillary"):
    """Build a list of integer labels that are contained into a candidate
    filtered image; according to its pixels

    Parameters
    ----------
    filtered_image : np.array
        Image to label, under the numpy.array format
    label_ids : list
        List of labels ids contained into the reference classification

    Returns
    -------
    dict
        label ids occur or not in the image
    """
    image_data = np.array(filtered_image)
    available_labels = np.unique(image_data)
    if dataset == "aerial":
        available_labels[available_labels == 255] = 1
    return {i: 1 if i in available_labels else 0 for i in label_ids}


def resize_image(img, base_size):
    """ Resize image `img` such that min(width, height)=base_size; keep image
    proportions

    Parameters:
    -----------
    img: Image
        input image to resize
    base_size: integer
        minimal dimension of the returned image
    """
    old_width, old_height = img.size
    if old_width < old_height:
        new_size = (base_size, int(base_size * old_height / old_width))
    else:
        new_size = (int(base_size * old_width / old_height), base_size)
    return img.resize(new_size)


def mono_crop_image(img, crop_pixel):
    """Crop image `img` so as its dimensions become equal (width=height),
    without modifying its smallest dimension
    """
    if img.width > img.height:
        return img.crop((crop_pixel, 0, crop_pixel + img.height, img.height))
    else:
        return img.crop((0, crop_pixel, img.width, crop_pixel + img.width))


def flip_image(img, proba=0.5):
    """ Flip image `img` horizontally with a probability of `proba`

    Parameters:
    -----------
    img: Image
        input image to resize
    proba: float
        probability of flipping input image (if less than 0.5, no flipping)
    """
    if np.random.sample() < proba:
        return img
    else:
        return img.transpose(Image.FLIP_LEFT_RIGHT)


def list_to_str(seq, sep="-"):
    """Transform the input sequence into a ready-to-print string

    Parameters
    ----------
    seq : list, tuple, dict
        Input sequence that must be transformed
    sep : str
        Separator that must appears between each `seq` items

    Returns
    -------
    str
        Printable version of input list
    """
    return sep.join(str(i) for i in seq)


def prepare_input_folder(datapath, dataset):
    """Data path and repository management; create and return the raw dataset
    path

    Parameters
    ----------
    datapath : str
        Data root directory, contain all used the datasets
    dataset : str
        Dataset name, *e.g.* `mapillary` or `shapes`

    Returns
    -------
    str
        Dataset raw image path

    """
    input_folder = os.path.join(datapath, dataset, "input")
    os.makedirs(input_folder, exist_ok=True)
    return input_folder


def prepare_preprocessed_folder(
    datapath, dataset, image_size
):
    """Data path and repository management; create all the folders needed to
    accomplish the current instance training/testing

    Parameters
    ----------
    datapath : str
        Data root directory, contain all used the datasets
    dataset : str
        Dataset name, *e.g.* `mapillary` or `shapes`
    image_size : int
        Size of the considered images (height and width are equal)

    Returns
    -------
    dict
        All the meaningful folders and dataset configuration file as a
    dictionary
    """
    prepro_folder = os.path.join(
        datapath,
        dataset,
        "preprocessed",
        str(image_size)
    )
    training_filename = os.path.join(prepro_folder, "training.json")
    validation_filename = os.path.join(prepro_folder, "validation.json")
    testing_filename = os.path.join(prepro_folder, "testing.json")
    training_folder = os.path.join(prepro_folder, "training")
    validation_folder = os.path.join(prepro_folder, "validation")
    testing_folder = os.path.join(prepro_folder, "testing")
    os.makedirs(os.path.join(training_folder, "images"), exist_ok=True)
    os.makedirs(os.path.join(training_folder, "labels"), exist_ok=True)
    os.makedirs(os.path.join(validation_folder, "images"), exist_ok=True)
    os.makedirs(os.path.join(validation_folder, "labels"), exist_ok=True)
    os.makedirs(os.path.join(testing_folder, "images"), exist_ok=True)
    return {
        "training": training_folder,
        "validation": validation_folder,
        "testing": testing_folder,
        "training_config": training_filename,
        "validation_config": validation_filename,
        "testing_config": testing_filename,
    }


def prepare_output_folder(datapath, dataset, image_size, model):
    """Dataset and repository management; create and return the dataset output
    path

    Parameters
    ----------
    datapath : str
        Data root directory, contain all used the datasets
    dataset : str
        Dataset name, *e.g.* `mapillary` or `shapes`
    image_size : int
        Size of the considered images (height and width are equal)
    model : str
        Research problem that is tackled, *e.g.* `featdet` or `semseg`

    Returns
    -------
    dict
        Dataset output paths
    """
    output_folder = os.path.join(datapath, dataset, "output", model)
    os.makedirs(output_folder, exist_ok=True)
    checkpoint_folder = os.path.join(output_folder, "checkpoints")
    os.makedirs(checkpoint_folder, exist_ok=True)
    best_model_filename = "best-model-" + str(image_size) + ".h5"
    best_instance_filename = "best-instance-" + str(image_size) + ".json"
    label_folder = os.path.join(output_folder, "predicted_labels")
    os.makedirs(label_folder, exist_ok=True)
    geometry_folder = raster_folder = None
    if dataset in GEOGRAPHIC_DATASETS:
        geometry_folder = os.path.join(output_folder, "predicted_geometries")
        raster_folder = os.path.join(output_folder, "predicted_rasters")
        os.makedirs(geometry_folder, exist_ok=True)
        os.makedirs(raster_folder, exist_ok=True)
    return {
        "best-model": os.path.join(checkpoint_folder, best_model_filename),
        "best-instance": os.path.join(checkpoint_folder, best_instance_filename),
        "checkpoints": checkpoint_folder,
        "geometries": geometry_folder,
        "labels": label_folder,
        "rasters": raster_folder,
    }


def recover_instance(instance_path):
    """Recover instance specification for model design, *i.e.* dropout rate and
    network architecture

    Parameters
    ----------
    instance_path : str
        Path of the instance specifications on the file system

    Returns
    -------
    tuple
        Instance dropout and network architecture
    """
    with open(instance_path) as fobj:
        instance = json.load(fobj)
    return instance["dropout"], instance["network"]


def GetHTMLColor(color):
    """Convert single-channeled colors or (R, G, B) tuples to #RRGGBB.

    If the color has only one channel (grayscale, typically), it is converted
    to a (R, G, B) tuple first.

    Parameters
    ----------
    color : int or list
        Grayscale value or list of red, green, blue pixel values

    Returns
    -------
    str
        HTML-version of color
    """
    rgb_tuple = (color, color, color) if type(color) == int else color
    return "#%02x%02x%02x" % tuple(rgb_tuple)


def build_image_from_config(data, config):
    """Rebuild a labelled image version from dataset configuration

    Parameters
    ----------
    data : numpy.array
        Labelled image as a numpy array of shape (img_size, img_size, 3) ;
    squared-size, shape (imsize, imsize)
    config : dict
        Label definition and description

    Returns
    -------
    np.array
        RGB-version of labelled images, with shape (imsize, imsize, 3)
    """
    result = np.zeros(
        shape=(data.shape[0], data.shape[1], 3), dtype=data.dtype
    )
    for label in range(len(config)):
        mask = data == label
        result[mask] = config[label]["color"]
    return Image.fromarray(result)


def create_symlink(link_name, directory):
    """Create a symbolic link

    Based on
    https://github.com/elemoine/dotfiles/blob/master/install-dotfiles.py

    Parameters
    ----------
    link_name : string
        name of the symbolic link
    directory : string
        Targetted destination for the symbolic link
    """
    if os.path.islink(link_name):
        os.unlink(link_name)
    elif os.path.isfile(link_name):
        logger.error("%s is a file!", link_name)
    elif os.path.isdir(link_name):
        logger.error("%s is a directory!", link_name)
    os.symlink(directory, link_name)
    logger.info("%s now points to %s.", link_name, directory)



"""Design a semantic segmentation model with Keras API
"""

import daiquiri
import keras as K

from deeposlandia.network import ConvolutionalNeuralNetwork


logger = daiquiri.getLogger(__name__)


class SemanticSegmentationNetwork(ConvolutionalNeuralNetwork):
    """Class that encapsulates semantic segmentation network creation

    Inherits from `ConvolutionalNeuralNetwork`

    Attributes
    ----------
    network_name : str
        Name of the network
    image_size : integer
        Input image size (height and width are equal)
    nb_channels : integer
        Number of input image channels (1 for greyscaled images, 3 for RGB
    images)
    nb_labels : integer
        Number of classes in the dataset glossary
    X : tensor
        (batch_size, image_size, image_size, nb_channels)-shaped input tensor;
    input image data
    Y : tensor
        (batch_size, image_size, image_size, nb_classes)-shaped output tensor,
    built as the output of the last network layer
    """

    def __init__(
        self,
        network_name="mapillary",
        image_size=512,
        nb_channels=3,
        nb_labels=65,
        dropout=1.0,
        architecture="simple",
    ):
        super().__init__(
            network_name, image_size, nb_channels, nb_labels, dropout
        )
        if architecture == "unet":
            error_msg = (
                "Please consider a divisible-per-16 image size "
                "for this architecture."
            )
            assert image_size % 16 == 0, error_msg
            self.Y = self.unet()
        elif architecture == "dilated":
            self.Y = self.dilated(add_context=True)
        elif architecture == "simple":
            self.Y = self.simple()
        else:
            logger.error(
                (
                    "Unknown network architecture. Please use "
                    "'simple' or 'unet'."
                )
            )
            raise ValueError("Unknown network architecture.")

    def output_layer(self, x, depth):
        """Build an output layer to a neural network, as a dense layer with
        sigmoid activation function as the point is to detect multiple labels
        on a single image

        Parameters
        ----------
        x : tensor
            Previous layer within the neural network (last hidden layer)
        depth : integer
            Dimension of the previous neural network layer

        Returns
        -------
        tensor
            2D output layer
        """
        y = K.layers.Conv2DTranspose(
            depth, kernel_size=2, padding="same", name="output_trconv"
        )(x)
        y = K.layers.Activation("softmax", name="output_activation")(y)
        return y

    def simple(self):
        """Build a simple default convolutional neural network composed of 3
        convolution-maxpool blocks and 1 fully-connected layer

        Returns
        -------
        tensor
            (batch_size, image_size, image_size, nb_labels)-shaped output
        predictions, that have to be compared with ground-truth values
        """
        layer = self.convolution(
            self.X, nb_filters=32, kernel_size=3, block_name="conv1"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool1")
        layer = self.convolution(
            layer, nb_filters=64, kernel_size=3, block_name="conv2"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool2")
        layer = self.convolution(
            layer, nb_filters=128, kernel_size=3, block_name="conv3"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool3")
        layer = self.transposed_convolution(
            layer,
            nb_filters=128,
            strides=2,
            kernel_size=3,
            block_name="trconv1",
        )
        layer = self.transposed_convolution(
            layer,
            nb_filters=64,
            strides=2,
            kernel_size=3,
            block_name="trconv2",
        )
        layer = self.transposed_convolution(
            layer,
            nb_filters=32,
            strides=2,
            kernel_size=3,
            block_name="trconv3",
        )
        return self.output_layer(layer, depth=self.nb_labels)

    def unet(self):
        """Build a U-net convolutional neural network; this architecture is
        characterized by a first set of convolution layer groups, each of which
        being followed by a max pooling layer so as to decrease the image size,
        and by a second set of convolution layer groups that are sprinkled by
        upsampling layers. A link between the first decreasing part and the
        second increasing part is ensured by crosswise concatenation
        operations.

        See: Ronneberger, Fischer & Brox, U-Net: Convolutional Networks for
        Biomedical Image Segmentation, arXiv technical report, 2015

        Returns
        -------
        tensor
            (batch_size, image_size, image_size, nb_labels)-shaped output
        predictions, that have to be compared with ground-truth values

        """
        conv1 = self.convolution(
            self.X, nb_filters=32, kernel_size=3, block_name="conv1a"
        )
        conv1 = self.convolution(
            conv1, nb_filters=32, kernel_size=3, block_name="conv1b"
        )

        pool1 = self.maxpool(conv1, pool_size=2, strides=2, block_name="pool1")
        conv2 = self.convolution(
            pool1, nb_filters=64, kernel_size=3, block_name="conv2a"
        )
        conv2 = self.convolution(
            conv2, nb_filters=64, kernel_size=3, block_name="conv2b"
        )

        pool2 = self.maxpool(conv2, pool_size=2, strides=2, block_name="pool2")
        conv3 = self.convolution(
            pool2, nb_filters=128, kernel_size=3, block_name="conv3a"
        )
        conv3 = self.convolution(
            conv3, nb_filters=128, kernel_size=3, block_name="conv3b"
        )

        pool3 = self.maxpool(conv3, pool_size=2, strides=2, block_name="pool3")
        conv4 = self.convolution(
            pool3, nb_filters=256, kernel_size=3, block_name="conv4a"
        )
        conv4 = self.convolution(
            conv4, nb_filters=256, kernel_size=3, block_name="conv4b"
        )

        pool4 = self.maxpool(conv4, pool_size=2, strides=2, block_name="pool4")
        conv5 = self.convolution(
            pool4, nb_filters=512, kernel_size=3, block_name="conv5a"
        )
        conv5 = self.convolution(
            conv5, nb_filters=512, kernel_size=3, block_name="conv5b"
        )

        up1 = self.upsample(conv5, conv4, block_name="up1")
        conv6 = self.convolution(
            up1, nb_filters=256, kernel_size=3, block_name="conv6a"
        )
        conv6 = self.convolution(
            conv6, nb_filters=256, kernel_size=3, block_name="conv6b"
        )

        up2 = self.upsample(conv6, conv3, block_name="up2")
        conv7 = self.convolution(
            up2, nb_filters=128, kernel_size=3, block_name="conv7a"
        )
        conv7 = self.convolution(
            conv7, nb_filters=128, kernel_size=3, block_name="conv7b"
        )

        up3 = self.upsample(conv7, conv2, block_name="up3")
        conv8 = self.convolution(
            up3, nb_filters=64, kernel_size=3, block_name="conv8a"
        )
        conv8 = self.convolution(
            conv8, nb_filters=64, kernel_size=3, block_name="conv8b"
        )

        up4 = self.upsample(conv8, conv1, block_name="up4")
        conv9 = self.convolution(
            up4, nb_filters=32, kernel_size=3, block_name="conv9a"
        )
        conv9 = self.convolution(
            conv9, nb_filters=32, kernel_size=3, block_name="conv9b"
        )
        conv9 = self.convolution(
            conv9,
            nb_filters=self.nb_labels,
            kernel_size=1,
            block_name="conv9c",
        )
        return K.layers.Activation("softmax", name="output_activation")(conv9)

    def dilated(self, add_context=False):
        """Build a dilated convolution network with two parts: a front-end part
        which is essentially an adapted VGG network, and a context part that
        corresponds to a sequence of layers with dilated convolutions.

        See: Yu, Koltun (2016). Multi-scale context aggregation by dilated
        convolutions, ArXiV technical report. Available at
        https://arxiv.org/pdf/1511.07122.pdf

        Parameters
        ----------
        add_context : boolean
             If true, the context block is added to the architecture

        Returns
        -------
        tensor
            (batch_size, image_size, image_size, nb_labels)-shaped output
        predictions, that have to be compared with ground-truth values

        """
        conv1 = self.convolution(
            self.X,
            nb_filters=64,
            kernel_size=3,
            batch_norm=False,
            block_name="conv1a_fe",
        )
        conv1 = self.convolution(
            conv1,
            nb_filters=64,
            kernel_size=3,
            batch_norm=False,
            block_name="conv1b_fe",
        )
        pool1 = self.maxpool(
            conv1, pool_size=2, strides=2, block_name="pool1_fe"
        )
        conv2 = self.convolution(
            pool1,
            nb_filters=128,
            kernel_size=3,
            batch_norm=False,
            block_name="conv2a_fe",
        )
        conv2 = self.convolution(
            conv2,
            nb_filters=128,
            kernel_size=3,
            batch_norm=False,
            block_name="conv2b_fe",
        )
        pool2 = self.maxpool(
            conv2, pool_size=2, strides=2, block_name="pool2_fe"
        )

        conv3 = self.convolution(
            pool2,
            nb_filters=256,
            kernel_size=3,
            batch_norm=False,
            block_name="conv3a_fe",
        )
        conv3 = self.convolution(
            conv3,
            nb_filters=256,
            kernel_size=3,
            batch_norm=False,
            block_name="conv3b_fe",
        )
        conv3 = self.convolution(
            conv3,
            nb_filters=256,
            kernel_size=3,
            batch_norm=False,
            block_name="conv3c_fe",
        )
        pool3 = self.maxpool(
            conv3, pool_size=2, strides=2, block_name="pool3_fe"
        )

        conv4 = self.convolution(
            pool3,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            block_name="conv4a_fe",
        )
        conv4 = self.convolution(
            conv4,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            block_name="conv4b_fe",
        )
        conv4 = self.convolution(
            conv4,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            block_name="conv4c_fe",
        )

        conv5 = self.convolution(
            conv4,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            dilation_rate=2,
            block_name="conv5a_fe",
        )
        conv5 = self.convolution(
            conv5,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            dilation_rate=2,
            block_name="conv5b_fe",
        )
        conv5 = self.convolution(
            conv5,
            nb_filters=512,
            kernel_size=3,
            batch_norm=False,
            dilation_rate=2,
            block_name="conv5c_fe",
        )

        conv6 = self.convolution(
            conv5,
            nb_filters=4096,
            kernel_size=7,
            batch_norm=False,
            dilation_rate=4,
            block_name="conv6_fe",
        )
        conv6 = K.layers.Dropout(0.5, name="do1")(conv6)

        conv7 = self.convolution(
            conv6,
            nb_filters=4096,
            kernel_size=1,
            batch_norm=False,
            block_name="conv7_fe",
        )
        conv7 = K.layers.Dropout(0.5, name="do2")(conv7)

        conv8 = self.convolution(
            conv7,
            nb_filters=self.nb_labels,
            batch_norm=False,
            kernel_size=1,
            activation="linear",
            block_name="conv8_fe",
        )

        if add_context:
            dilated_net = self.add_dilated_context(conv8)
        else:
            dilated_net = conv8

        # upsample
        net = self.transposed_convolution(
            dilated_net,
            nb_filters=256,
            batch_norm=False,
            kernel_size=3,
            strides=2,
            block_name="trconv1",
        )
        net = self.transposed_convolution(
            net,
            nb_filters=256,
            batch_norm=False,
            kernel_size=3,
            strides=2,
            block_name="trconv2",
        )
        net = self.transposed_convolution(
            net,
            nb_filters=self.nb_labels,
            batch_norm=False,
            activation="softmax",
            kernel_size=3,
            strides=2,
            block_name="trconv3",
        )
        return net



"""Define a class that represents typical convolutional neural networks
"""

import keras as K


class ConvolutionalNeuralNetwork:
    """Convolutional neural network design

    Attributes
    ----------
    network_name : str
        Name of the network
    image_size : integer
        Input image size (height and width are equal)
    nb_channels : integer
        Number of input image channels (1 for greyscaled images, 3 for RGB
    images)
    nb_labels : integer
        Number of classes in the dataset glossary
    X : tensor
        (batch_size, image_size, image_size, nb_channels)-shaped input tensor;
    input image data
    """

    def __init__(
        self,
        network_name="mapillary",
        image_size=224,
        nb_channels=3,
        nb_labels=65,
        dropout=1.0,
    ):
        if not image_size % 16 == 0:
            raise ValueError(
                "The chosen image size is not divisible "
                "per 16. To train a neural network with "
                "such an input size may fail."
            )
        self.network_name = network_name
        self.image_size = image_size
        self.nb_channels = nb_channels
        self.nb_labels = nb_labels
        self.dropout_rate = dropout
        self.X = K.layers.Input(
            shape=(image_size, image_size, nb_channels), name="input"
        )

    def layer_name(self, prefix, suffix):
        """Concatenate prefix and suffix to build a complete layer name

        Use the default Keras behavior if prefix is None

        Parameters
        ----------
        prefix : str
            Layer name prefix, refers to the layer block
        suffix : str
            Layer name suffix, refers to the layer type

        Returns
        -------
        str
            Complete layer name, build as prefix + suffix
        """
        return prefix + suffix if prefix is not None else None

    def convolution(
        self,
        x,
        nb_filters,
        kernel_size,
        strides=1,
        dilation_rate=1,
        padding="same",
        activation="relu",
        batch_norm=True,
        block_name=None,
    ):
        """Apply a convolutional layer within a neural network

        Use Keras API

        Parameters
        ----------
        x : tensor
            Input layer
        nb_filters : integer
            Number of convolution filters
        kernel_size : integer
            Convolution filter size, in pixel
        strides : integer
            Convolution strides, in pixel
        dilation_rate : integer
            Rate of dilation, for atrous convolution (default to 1, no
        dilation)
        padding : str
            Border pixel management ("valid" to apply convolution pixel only on
        image pixels, or "same" to replicate border pixels)
        activation : str
            Type of activation function to apply on the tensor at the end of
        the convolution block ('relu' by default)
        batch_norm : boolean
            If True, a batch normalization process is applied on `x` tensor
        before activation layer
        block_name : str
            Convolution block name, for identification purpose

        Returns
        -------
        tensor
            4D output layer
        """
        x = K.layers.Conv2D(
            nb_filters,
            kernel_size=kernel_size,
            strides=strides,
            dilation_rate=dilation_rate,
            padding=padding,
            name=self.layer_name(block_name, "_conv"),
        )(x)
        if batch_norm:
            x = K.layers.BatchNormalization(
                name=self.layer_name(block_name, "_bn")
            )(x)
        x = K.layers.Activation(
            activation, name=self.layer_name(block_name, "_activation")
        )(x)
        return x

    def transposed_convolution(
        self,
        x,
        nb_filters,
        kernel_size,
        strides=1,
        padding="same",
        activation="relu",
        batch_norm=True,
        block_name=None,
    ):
        """Build a layer seen as the transpose operation of classic
        convolution, for a convolutional neural network

        Use Keras API

        Parameters
        ----------
        x : tensor
            Input tensor
        nb_filters : integer
            Number of convolution filters
        kernel_size : integer
            Convolution filter size, in pixel
        strides : integer
            Convolution strides, in pixel
        padding : str
            Border pixel management ("valid" to apply convolution pixel only on
        image pixels, or "same" to replicate border pixels)
        activation : str
            Type of activation function to apply on the tensor at the end of
        the convolution block ('relu' by default)
        batch_norm : boolean
            If True, a batch normalization process is applied on `x` tensor
        before activation layer
        block_name : str
            Transposed convolution block name, for identification purpose

        Returns
        -------
        tensor
            4D output layer
        """
        x = K.layers.Conv2DTranspose(
            nb_filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            name=self.layer_name(block_name, "_transconv"),
        )(x)
        if batch_norm:
            x = K.layers.BatchNormalization(
                name=self.layer_name(block_name, "_bn")
            )(x)
        x = K.layers.Activation(
            activation, name=self.layer_name(block_name, "_activation")
        )(x)
        return x

    def maxpool(
        self, x, pool_size, strides=1, padding="same", block_name=None
    ):
        """Apply a max pooling layer within a neural network

        Use Keras API

        Parameters
        ----------
        x : tensor
            Input layer
        pool_size : integer
            Pooling kernel size, in pixel
        strides : integer
            Pooling strides, in pixel ; indicates the downscaling factor
        padding : str
            Border pixel management ("valid" to apply convolution pixel only on
        image pixels, or "same" to replicate border pixels)
        block_name : str
            Pooling block name, for identification purpose

        Returns
        -------
        tensor
            4D output layer
        """
        return K.layers.MaxPool2D(
            pool_size=pool_size,
            strides=strides,
            padding=padding,
            name=block_name,
        )(x)

    def dense(
        self, x, depth, activation="relu", batch_norm=True, block_name=None
    ):
        """Apply a fully-connected layer within a neural network

        Use Keras API

        Parameters
        ----------
        x : tensor
            Input layer
        depth : integer
            Number of neurons used within the layer
        activation : str
            Type of activation function to apply on the tensor at the end of
        the convolution block ('relu' by default)
        batch_norm : boolean
            If True, a batch normalization process is applied on `x` tensor
        before activation layer
        block_name : str
            Fully-connected block name, for identification purpose

        Returns
        -------
        tensor
            Output layer
        """
        x = K.layers.Dense(depth, name=self.layer_name(block_name, "_fc"))(x)
        if batch_norm:
            x = K.layers.BatchNormalization(
                name=self.layer_name(block_name, "_bn")
            )(x)
        x = K.layers.Activation(
            activation, name=self.layer_name(block_name, "_activation")
        )(x)
        x = K.layers.Dropout(
            self.dropout_rate, name=self.layer_name(block_name, "_dropout")
        )(x)
        return x

    def flatten(self, x, block_name=None):
        """Apply a flattening operation to input tensor `x`, to reduce its
        dimension; arises generally before a dense layer

        Parameters
        ----------
        x : tensor
            Input layer; its shapes is necessarily larger than 3
        block_name : str
            Flatten block name, for identification purpose

        Returns
        -------
        tensor
            2D output layer
        """
        return K.layers.Flatten(name=block_name)(x)

    def upsample(self, layer1, layer2, block_name=None):
        """Apply an upsampling operation on `layer1` and concatenate the
        resulting layer with `layer2`

        Parameters
        ----------
        layer1 : tensor
            First input layer, its shape must correspond to layer2 shape
        layer2 : tensor
            Second input layer, its shape must correspond to layer1 shape
        block_name : str
            Upsample block name, for identification purpose

        """
        upname = self.layer_name(block_name, "_up")
        ccname = self.layer_name(block_name, "_concat")
        upsample = K.layers.UpSampling2D(size=(2, 2), name=upname)(layer1)
        return K.layers.concatenate([upsample, layer2], axis=3, name=ccname)

    def add_dilated_context(self, input_layer):
        """Add a context block that corresponds to Yu et al. (2016)
        contribution, in order to aggregate multi-scale contextual information

        Parameters
        ----------
        input_layer : tensor
            Input layer, before to add multi-scale context

        """
        context = K.layers.ZeroPadding2D(33)(input_layer)
        context = self.convolution(
            context,
            nb_filters=2 * self.nb_labels,
            kernel_size=3,
            dilation_rate=1,
            padding="valid",
            block_name="conv1_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=2 * self.nb_labels,
            kernel_size=3,
            dilation_rate=1,
            padding="valid",
            block_name="conv2_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=4 * self.nb_labels,
            kernel_size=3,
            dilation_rate=2,
            padding="valid",
            block_name="conv3_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=8 * self.nb_labels,
            kernel_size=3,
            dilation_rate=4,
            padding="valid",
            block_name="conv4_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=16 * self.nb_labels,
            kernel_size=3,
            dilation_rate=8,
            padding="valid",
            block_name="conv5_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=32 * self.nb_labels,
            kernel_size=3,
            dilation_rate=16,
            padding="valid",
            block_name="conv6_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=32 * self.nb_labels,
            kernel_size=3,
            dilation_rate=1,
            padding="valid",
            block_name="conv7_ctx",
        )
        context = self.convolution(
            context,
            nb_filters=self.nb_labels,
            kernel_size=1,
            dilation_rate=1,
            padding="valid",
            activation="linear",
            block_name="conv8_ctx",
        )
        return context



"""Set of functions dedicated to georeferenced object handling

The functions which allow to convert raster in vector and vice-versa are
inspired from:
  - https://www.kaggle.com/lopuhin/full-pipeline-demo-poly-pixels-ml-poly

"""

from collections import defaultdict

import cv2
import daiquiri
from fiona.crs import from_epsg
import geopandas as gpd
import numpy as np
import shapely.geometry as shgeom


logger = daiquiri.getLogger(__name__)


def get_pixel(coord, min_coord, max_coord, size):
    """Transform abscissa from geographical coordinate to pixel

    For horizontal operations, 'min_coord', 'max_coord' and 'size' refer
    respectively to west and east coordinates and image width.

    For vertical operations, 'min_coord', 'max_coord' and 'size' refer
    respectively to north and south coordinates and image height.

    Parameters
    ----------
    coord : list
        Coordinates to transform
    min_coord : float
        Georeferenced minimal coordinate of the image
    max_coord : float
        Georeferenced maximal coordinate of the image
    size : int
        Image size, in pixels

    Returns
    -------
    list
        Transformed coordinates, as pixel references within the image
    """
    if isinstance(coord, list):
        return [
            int(size * (c - min_coord) / (max_coord - min_coord))
            for c in coord
        ]
    elif isinstance(coord, float):
        return int(size * (coord - min_coord) / (max_coord - min_coord))
    else:
        raise TypeError(
            "Unknown type (%s), pass a 'list' or a 'float'", type(coord)
        )


def get_geocoord(coord, min_coord, max_coord, size):
    """Transform abscissa from pixel to geographical coordinate

    For horizontal operations, 'min_coord', 'max_coord' and 'size' refer
    respectively to west and east coordinates and image width.

    For vertical operations, 'min_coord', 'max_coord' and 'size' refer
    respectively to north and south coordinates and image height.

    Parameters
    ----------
    coord : list
        Coordinates to transform
    min_coord : float
        Minimal coordinates of the image, as a pixel reference
    max_coord : float
        Maximal coordinates of the image, as a pixel reference
    size : int
        Image size, in pixels

    Returns
    -------
    list
        Transformed coordinates, expressed in the accurate coordinate system
    """
    if isinstance(coord, list):
        return [min_coord + c * (max_coord - min_coord) / size for c in coord]
    elif isinstance(coord, int):
        return min_coord + coord * (max_coord - min_coord) / size
    else:
        raise TypeError(
            "Unknown type (%s), pass a 'list' or a 'int'", type(coord)
        )


def get_image_features(raster):
    """Retrieve geotiff image features with GDAL

    Use the `GetGeoTransform` method, that provides the following values:
        + East/West location of Upper Left corner
        + East/West pixel resolution
        + 0.0
        + North/South location of Upper Left corner
        + 0.0
        + North/South pixel resolution

    See GDAL documentation (https://www.gdal.org/gdal_tutorial.html)

    Parameters
    ----------
    raster : osgeo.gdal.Dataset
        Active opened image as a GDAL object

    Returns
    -------
    dict
        Bounding box of the image (west, south, east, north coordinates), srid,
        and size (in pixels)

    """
    width = raster.RasterXSize
    height = raster.RasterYSize
    gt = raster.GetGeoTransform()
    minx = gt[0]
    miny = gt[3] + height * gt[5]
    maxx = gt[0] + width * gt[1]
    maxy = gt[3]
    srid = int(raster.GetProjection().split('"')[-2])
    return {
        "west": minx,
        "south": miny,
        "east": maxx,
        "north": maxy,
        "srid": srid,
        "width": width,
        "height": height,
    }


def extract_points_from_polygon(p, geofeatures, min_x, min_y):
    """Extract points from a polygon

    Parameters
    ----------
    p : shapely.geometry.Polygon
        Polygon to detail
    geofeatures : dict
        Geographical features associated to the image
    min_x : int
        Minimal x-coordinate (west)
    min_y : int
        Minimal y-coordinate (north)
    Returns
    -------
    numpy.array
        Polygon vertices

    """
    raw_xs, raw_ys = p.exterior.xy
    xs = get_pixel(
        list(raw_xs),
        geofeatures["west"],
        geofeatures["east"],
        geofeatures["width"],
    )
    ys = get_pixel(
        list(raw_ys),
        geofeatures["north"],
        geofeatures["south"],
        geofeatures["height"],
    )
    points = np.array([[y, x] for x, y in zip(xs, ys)], dtype=np.int32)
    points[:, 0] -= min_y
    points[:, 1] -= min_x
    return points


def get_tile_footprint(features, min_x, min_y, tile_width, tile_height=None):
    """Compute a tile geographical footprint expressed as a `shapely` geometry
    that contains geographical coordinates of tile corners

    Parameters
    ----------
    features : dict
        Raw image raster geographical features (`north`, `south`, `east` and
    `west` coordinates, `weight` and `height` measured in pixels)
    min_x : int
        Left tile limit, as a horizontal pixel index
    min_y : int
        Upper tile limit, as a vertical pixel index
    tile_width : int
        Tile width, measured in pixel
    tile_height : int
        Tile height, measured in pixel: if None, consider
    `tile_height=tile_width` (squared tile)

    Returns
    -------
    shapely.geometry.Polygon
        Tile footprint, as a square polygon delimited by its corner
    geographical coordinates

    """
    tile_height = tile_width if tile_height is None else tile_height
    max_x = min_x + tile_width
    max_y = min_y + tile_height
    min_x_coord, max_x_coord = get_geocoord(
        [min_x, max_x], features["west"], features["east"], features["width"]
    )
    min_y_coord, max_y_coord = get_geocoord(
        [min_y, max_y],
        features["north"],
        features["south"],
        features["height"],
    )
    return shgeom.Polygon(
        (
            (min_x_coord, min_y_coord),
            (max_x_coord, min_y_coord),
            (max_x_coord, max_y_coord),
            (min_x_coord, max_y_coord),
        )
    )


def extract_tile_items(
    raster_features, labels, min_x, min_y, tile_width, tile_height
):
    """Extract label items that belong to the tile defined by the minimum
    horizontal pixel `min_x` (left tile limit), the minimum vertical pixel
    `min_y` (upper tile limit) and the sizes ̀tile_width` and `tile_height`
    measured as a pixel amount.

    The tile is cropped from the original image raster as follows:
      - horizontally, between `min_x` and `min_x+tile_width`
      - vertically, between `min_y` and `min_y+tile_height`

    This method takes care of original data projection (UTM 37S, Tanzania
    area), however this parameter may be changed if similar data on another
    projection is used.

    Parameters
    ----------
    raster_features : dict
        Raw image raster geographical features (`north`, `south`, `east` and
    `west` coordinates, `weight` and `height` measured in pixels)
    labels : geopandas.GeoDataFrame
        Raw image labels, as a set of geometries
    min_x : int
        Left tile limit, as a horizontal pixel index
    min_y : int
        Upper tile limit, as a vertical pixel index
    tile_width : int
        Tile width, measured in pixel
    tile_height : int
        Tile height, measured in pixel

    Returns
    -------
    geopandas.GeoDataFrame
        Set of ground-truth labels contained into the tile, characterized by
    their type (complete, unfinished or foundation) and their geometry

    """
    area = get_tile_footprint(
        raster_features, min_x, min_y, tile_width, tile_height
    )
    bdf = gpd.GeoDataFrame(
        crs=from_epsg(raster_features["srid"]), geometry=[area]
    )
    reproj_labels = labels.to_crs(epsg=raster_features["srid"])
    tile_items = gpd.sjoin(reproj_labels, bdf)
    if tile_items.shape[0] == 0:
        return tile_items[["condition", "geometry"]]
    tile_items = gpd.overlay(tile_items, bdf)
    tile_items = tile_items.explode()  # Manage MultiPolygons
    return tile_items[["condition", "geometry"]]


def extract_geometry_vertices(mask, structure_size=(10, 10), approx_eps=0.01):
    """Extract polygon vertices from a boolean mask with the help of OpenCV
    utilities, as a numpy array

    Parameters
    ----------
    mask : numpy.array
        Image mask where to find polygons
    structure_size : tuple
        Size of the cv2 structuring artefact, as a tuple of horizontal and
    vertical pixels
    approx_eps : double
        Approximation coefficient, aiming at building more simplified polygons
    (this coefficient lies between 0 and 1, the larger the value is, the more
    important the approximation is)

    Returns
    -------
    numpy.array
        List of polygons contained in the mask, identified by their vertices
    """
    structure = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, structure_size)
    denoised = cv2.morphologyEx(mask, cv2.MORPH_OPEN, structure)
    grown = cv2.morphologyEx(denoised, cv2.MORPH_CLOSE, structure)
    _, contours, hierarchy = cv2.findContours(
        grown, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
    )
    polygons = [
        cv2.approxPolyDP(
            c, epsilon=approx_eps * cv2.arcLength(c, closed=True), closed=True
        )
        for c in contours
    ]
    return polygons, hierarchy


def retrieve_area_color(data, contour, labels):
    """Mask an image area and retrieve its dominant color starting from a label
    glossary, by determining its closest label (regarding euclidean distance).

    Largely inspired from : https://www.pyimagesearch.com/\
    2016/02/15/determining-object-color-with-opencv/

    Parameters
    ----------
    data : np.array
        3-channelled image
    contour : np.array
        List of points that delimits the area
    labels : list
        List of dictionnary that describes each labels (with "id" and "color" keys)
    """
    mask = np.zeros(data.shape[:2], dtype="uint8")
    cv2.drawContours(mask, [contour], -1, 255, -1)
    mean = cv2.mean(data, mask=mask)[:3]
    min_dist = (np.inf, None)
    for label in labels:
        d = np.linalg.norm(label["color"] - np.array(mean))
        if d < min_dist[0]:
            min_dist = (d, label["id"])
    return min_dist[1]


def vectorize_mask(
    mask, colored_data, labels,
    min_area=10.0, structure_size=(10, 10), approx_eps=0.01
):
    """Convert a numpy array (*i.e.* rasterized information) into a shapely
    Multipolygon (*i.e.* vectorized information), by filtering too small
    objects. As a reminder, this function is supposed to be used for building
    detection, hence by definition a building can not be smaller than a few
    square meters

    A large part of the function (and comments) comes from:
      - https://www.kaggle.com/lopuhin/full-pipeline-demo-poly-pixels-ml-poly

    See also
    http://docs.opencv.org/3.1.0/d9/d8b/tutorial_py_contours_hierarchy.html

    Parameters
    ----------
    mask : numpy.array
        1-channelled version of the input image
    data : numpy.array
        3-channelled version of the input  image
    labels : list of dicts
        List of labels, as dicts that contain at least a "id"
    min_area : double
    structure_size : tuple
        Size of the cv2 structuring artefact, as a tuple of horizontal and
    vertical pixels
    approx_eps : double
        Approximation coefficient, aiming at building more simplified polygons
    (this coefficient lies between 0 and 1, the larger the value is, the more
    important the approximation is)

    Returns
    -------
    shapely.geometry.MultiPolygon
        Set of detected objects grouped as a MultiPolygon object
    """
    contours, hierarchy = extract_geometry_vertices(
        mask, structure_size, approx_eps
    )
    if not contours:
        return np.array([]), shgeom.MultiPolygon()
    # now messy stuff to associate parent and child contours
    cnt_children = defaultdict(list)
    child_contours = set()
    assert hierarchy.shape[0] == 1
    for idx, (_, _, _, parent_idx) in enumerate(hierarchy[0]):
        if parent_idx != -1:
            child_contours.add(idx)
            cnt_children[parent_idx].append(contours[idx])
    # create actual polygons filtering by area (removes artifacts)
    all_polygons = []
    out_labels = []
    for idx, cnt in enumerate(contours):
        if idx not in child_contours and cv2.contourArea(cnt) >= min_area:
            assert cnt.shape[1] == 1
            poly = shgeom.Polygon(
                shell=cnt[:, 0, :],
                holes=[
                    c[:, 0, :]
                    for c in cnt_children.get(idx, [])
                    if cv2.contourArea(c) >= min_area
                ],
            )
            poly_label = retrieve_area_color(colored_data, cnt, labels)
            all_polygons.append(poly)
            out_labels.append(poly_label)
    all_polygons = shgeom.MultiPolygon(all_polygons)
    # approximating polygons might have created invalid ones, fix them
    if not all_polygons.is_valid:
        all_polygons = all_polygons.buffer(0)
        # Sometimes buffer() converts a simple Multipolygon to just a Polygon,
        # need to keep it a Multi throughout
        if all_polygons.type == "Polygon":
            all_polygons = shgeom.MultiPolygon([all_polygons])
    return np.array(out_labels), all_polygons


def rasterize_polygons(polygons, labels, img_height, img_width):
    """Transform a vectorized information into a numpy mask for plotting
    purpose

    Inspired from:
      - https://www.kaggle.com/lopuhin/full-pipeline-demo-poly-pixels-ml-poly

    Parameters
    ----------
    polygons : shapely.geometry.MultiPolygon
        Set of detected objects, stored as a MultiPolygon
    labels : np.array
        List of corresponding labels that describes polygon classes
    img_height : int
        Image height, in pixels
    img_width : int
        Image width, in pixels

    Returns
    -------
    numpy.array
        Rasterized polygons
    """
    img_mask = np.zeros(shape=(img_height, img_width), dtype=np.uint8)
    if not polygons:
        return img_mask
    for polygon, label in zip(polygons, labels):
        exterior = np.array(polygon.exterior.coords).round().astype(np.int32)
        interiors = [
            np.array(pi.coords).round().astype(np.int32)
            for pi in polygon.interiors
        ]
        cv2.fillPoly(img_mask, [exterior], int(label))
        cv2.fillPoly(img_mask, interiors, 0)
    return img_mask


def pixel_to_geocoord(polygon_ring, geofeatures):
    """Transform point coordinates from pixel to geographical coordinates in
    the accurate geographical projection, *i.e.* the projection of original
    image indicated in the `geofeatures` dict

    Parameters
    ----------
    polygon_ring : shapely.geometry.polygon.LinearRing
        Sequence of points belonging to the object of interest
    geofeatures : dict
        Geographical characteristics of the original image

    Returns
    -------
    numpy.array
        Transformed sequence of points, expressed in geographical coordinates,
    regarding the SRID provided by `geofeatures["srid"]`

    """
    xpixels, ypixels = polygon_ring.xy
    px = get_geocoord(
        list(xpixels),
        geofeatures["west"],
        geofeatures["east"],
        geofeatures["width"],
    )
    py = get_geocoord(
        list(ypixels),
        geofeatures["north"],
        geofeatures["south"],
        geofeatures["height"],
    )
    return np.array([px, py]).T


def convert_to_geocoord(polygons, geofeatures):
    """Convert pixel-defined polygons into georeferenced polygon with the
    projection of the original raster, provided by `geofeatures["srid"]`

    Parameters
    ----------
    polygons : shapely.geometry.MultiPolygon or list of
    shapely.geometry.Polygons
        Detected polygons identified with their vertex pixels on the image
    geofeatures : dict
        Geographical features associated to the original image

    Returns
    -------
    shapely.geometry.MultiPolygon
        Georeferenced multipolygon that describes every object belonging to
    original image
    """
    geopolygons = []
    for polygon in polygons:
        exterior = pixel_to_geocoord(polygon.exterior, geofeatures)
        interiors = [
            pixel_to_geocoord(pi, geofeatures) for pi in polygon.interiors
        ]
        if len(interiors) > 0:
            poly = shgeom.Polygon(shell=exterior, holes=[i for i in interiors])
        else:
            poly = shgeom.Polygon(shell=exterior)
        geopolygons.append(poly)
    return geopolygons



"""Shape dataset module

Model randomly generated images with geometric shapes, in order to get a toy
dataset that will make the convolutional neural network tests easier.
"""

import os
import math

import cv2
import daiquiri
import numpy as np
from PIL import Image

from deeposlandia.datasets import Dataset


logger = daiquiri.getLogger(__name__)


class ShapeDataset(Dataset):
    """Dataset structure that gathers all information related to a
    randomly-generated shape Dataset

    In such a dataset, a set of images is generated with either a square, or a
    circle or a triangle, or two of them, or all of them. A random background
    color is applied, and shape color itself is also randomly generated. Each
    of these labels are characterized with a fixed color for comparison between
    ground truth and predictions: squares, circles and triangles will be
    respectively set as blue, red and green, whilst background will be set as
    light grey.

    Attributes
    ----------
    image_size : int
        Size of considered images (height=width), raw images will be resized
    during the preprocessing
    nb_labels : int
        Number of shape types that must be integrated into the dataset (only 1,
    2 and 3 are supported)

    """

    SQUARE = 0
    SQUARE_COLOR = (50, 50, 200)  # Blue
    CIRCLE = 1
    CIRCLE_COLOR = (200, 50, 50)  # Red
    TRIANGLE = 2
    TRIANGLE_COLOR = (50, 200, 50)  # Green
    BACKGROUND = 3
    BACKGROUND_COLOR = (200, 200, 200)  # Light grey

    def __init__(self, image_size):
        """ Class constructor
        """
        super().__init__(image_size)
        self.build_glossary()

    def build_glossary(self):
        """Read the shape glossary stored as a json file at the data
        repository root

        Parameters
        ----------
        nb_labels : integer
            Number of shape types (either 1, 2 or 3, warning if more)
        """
        self.add_label(self.SQUARE, "square", self.SQUARE_COLOR, True)
        self.add_label(self.CIRCLE, "circle", self.CIRCLE_COLOR, True)
        self.add_label(self.TRIANGLE, "triangle", self.TRIANGLE_COLOR, True)
        self.add_label(
            self.BACKGROUND, "background", self.BACKGROUND_COLOR, True
        )

    def generate_labels(self, nb_images):
        """ Generate random shape labels in order to prepare shape image
        generation; use numpy to generate random indices for each labels, these
        indices will be the positive examples; return a 2D-list

        Parameters
        ----------
        nb_images : integer
            Number of images to label in the dataset
        """
        raw_labels = [
            np.random.choice(
                np.arange(nb_images), int(nb_images / 2), replace=False
            )
            for i in range(self.get_nb_labels())
        ]
        labels = np.zeros([nb_images, self.get_nb_labels()], dtype=int)
        for i in range(self.get_nb_labels()):
            labels[raw_labels[i], i] = 1
        return [dict([(i, int(j)) for i, j in enumerate(l)]) for l in labels]

    def populate(
        self,
        output_dir=None,
        input_dir=None,
        nb_images=10000,
        nb_tiles_per_image=None,
        aggregate=False,
        labelling=True,
        buf=8,
        nb_processes=None
    ):
        """ Populate the dataset with images contained into `datadir` directory

        Parameters
        ----------
        output_dir : str
            Path of the directory where the preprocessed image must be saved
        input_dir : str
            Path of the directory that contains input images
        nb_images: integer
            Number of images that must be added in the dataset
        nb_tiles_per_image : integer
            Number of tiles that must be picked into the raw image (useless there, added
        for consistency)
        aggregate: bool
            Aggregate some labels into more generic ones, e.g. cars and bus
        into the vehicle label
        labelling: boolean
            Dummy parameter: in this dataset, labels are always generated, as
        images are drawed with them
        buf: integer
            Minimal number of pixels between shape base point and image borders
        nb_processes : int
            Number of processes on which to run the preprocessing (dummy
        parameter for "shapes" datasets)
        """
        if nb_tiles_per_image is not None:
            logger.warning("The ``nb_tiles_per_image`` parameter is useless, it will be ignored.")
        shape_gen = self.generate_labels(nb_images)
        for i, image_label in enumerate(shape_gen):
            bg_color = np.random.randint(0, 255, 3).tolist()
            shape_specs = []
            for l in image_label.items():
                if l:
                    shape_color = np.random.randint(0, 255, 3).tolist()
                    x, y = np.random.randint(
                        buf, self.image_size - buf - 1, 2
                    ).tolist()
                    shape_size = np.random.randint(buf, self.image_size // 4)
                    shape_specs.append([shape_color, x, y, shape_size])
                else:
                    shape_specs.append([None, None, None, None])
            self.add_image(i, bg_color, shape_specs, image_label)
            if output_dir is not None:
                self.draw_image(i, output_dir)

    def add_image(self, image_id, background, specifications, labels):
        """ Add a new image to the dataset with image id `image_id`; an image
        in the dataset is represented by an id, a list of shape specifications,
        a background color and a list of 0-1 labels (1 if the i-th label is on
        the image, 0 otherwise)

        Parameters
        ----------
        image_id : integer
            Id of the new image
        background : list
            List of three integer between 0 and 255 that designs the image
        background color
        specifications : list
            Image specifications, as a list of shapes (color, coordinates and
        size)
        labels : list
            List of 0-1 values, the i-th value being 1 if the i-th label is on
        the new image, 0 otherwise; the label list length correspond to the
        number of labels in the dataset
        """
        if image_id in self.image_info:
            logger.error(
                "Image %s already stored into the label set.", image_id
            )
            return None
        self.image_info.append(
            {
                "background": background,
                "shape_specs": specifications,
                "labels": labels,
            }
        )

    def draw_image(self, image_id, datapath):
        """Draws an image from the specifications of its shapes and saves it on
        the file system to `datapath`

        Save labels as mono-channel images on the file system by using the
        label ids

        Parameters
        ----------
        image_id : integer
            Image id
        datapath : str
            String that characterizes the repository in which images will be
        stored
        """
        image_info = self.image_info[image_id]

        image = np.ones([self.image_size, self.image_size, 3], dtype=np.uint8)
        image = image * np.array(image_info["background"], dtype=np.uint8)
        label = np.full(
            [self.image_size, self.image_size, 3],
            self.BACKGROUND_COLOR,
            dtype=np.uint8,
        )

        # Get the center x, y and the size s
        if image_info["labels"][self.SQUARE]:
            color, x, y, s = image_info["shape_specs"][self.SQUARE]
            color = tuple(map(int, color))
            image = cv2.rectangle(
                image, (x - s, y - s), (x + s, y + s), color, -1
            )
            label = cv2.rectangle(
                label, (x - s, y - s), (x + s, y + s), self.SQUARE_COLOR, -1
            )
        if image_info["labels"][self.CIRCLE]:
            color, x, y, s = image_info["shape_specs"][self.CIRCLE]
            color = tuple(map(int, color))
            image = cv2.circle(image, (x, y), s, color, -1)
            label = cv2.circle(label, (x, y), s, self.CIRCLE_COLOR, -1)
        if image_info["labels"][self.TRIANGLE]:
            color, x, y, s = image_info["shape_specs"][self.TRIANGLE]
            color = tuple(map(int, color))
            x, y, s = map(int, (x, y, s))
            points = np.array(
                [
                    [
                        (x, y - s),
                        (x - s / math.sin(math.radians(60)), y + s),
                        (x + s / math.sin(math.radians(60)), y + s),
                    ]
                ],
                dtype=np.int32,
            )
            image = cv2.fillPoly(image, points, color)
            label = cv2.fillPoly(label, points, self.TRIANGLE_COLOR)
        image_filename = os.path.join(
            datapath, "images", "shape_{:05}.png".format(image_id)
        )
        self.image_info[image_id]["image_filename"] = image_filename
        Image.fromarray(image).save(image_filename)
        label_filename = os.path.join(
            datapath, "labels", "shape_{:05}.png".format(image_id)
        )
        self.image_info[image_id]["label_filename"] = label_filename
        Image.fromarray(label).save(label_filename)



"""Inference on deep learning model previously trained

* Build the instance name
* Load one (or more) image(s) from the file system
* Load a trained model starting from the instance name
* Make label predictions on the test image(s)
* Produce a result: for instance, only predicted labels

Example of program call, that will infers labels on all files from
̀path_to_images/shapes_00000.png` to `path_to_images/shapes_00009.png`::

    python deeposlandia/inference.py -D shapes -i path_to_img/shapes_0000*.png

"""

import argparse
import glob
import os

import daiquiri
import numpy as np
from PIL import Image

from keras.models import Model
import keras.backend as K

from deeposlandia import utils
from deeposlandia.datasets import AVAILABLE_DATASETS
from deeposlandia.feature_detection import FeatureDetectionNetwork
from deeposlandia.semantic_segmentation import SemanticSegmentationNetwork


logger = daiquiri.getLogger(__name__)


def init_model(
    problem, instance_name, image_size, nb_labels, dropout, network
):
    """Initialize a convolutional neural network with Keras API starting from a
    set of parameters

    Parameters
    ----------
    problem : str
        Type of solved problem, either `featdet` or `semseg`
    instance_name : str
        Name of the instance, for identification purpose
    image_size : int
        Image size, in pixels (height=width)
    nb_labels : int
        Number of output labels
    dropout : float
        Dropout rate
    network : str
        Network architecture

    Returns
    -------
    keras.models.Model
        Convolutional neural network
    """
    K.clear_session()
    if problem == "featdet":
        net = FeatureDetectionNetwork(
            network_name=instance_name,
            image_size=image_size,
            nb_labels=nb_labels,
            dropout=dropout,
            architecture=network,
        )
    elif problem == "semseg":
        net = SemanticSegmentationNetwork(
            network_name=instance_name,
            image_size=image_size,
            nb_labels=nb_labels,
            dropout=dropout,
            architecture=network,
        )
    else:
        raise ValueError(
            "Unrecognized model. Please enter 'featdet' or 'semseg'."
        )
    return Model(net.X, net.Y)


def predict(
    filenames,
    dataset,
    problem,
    datapath="./data",
    name=None,
    network=None,
    batch_size=None,
    dropout=None,
    learning_rate=None,
    learning_rate_decay=None,
    output_dir="/tmp/deeposlandia/predicted",
):
    """Make label prediction on image indicated by ̀filename`, according to
    considered `problem`

    Parameters
    ----------
    filenames : str
        Name of the image files on the file system
    dataset : str
        Name of the dataset
    problem : str
        Name of the considered model, either `featdet` or `semseg`
    datapath : str
        Relative path of dataset repository
    name : str
        Name of the saved network
    network : str
        Name of the chosen architecture, either `simple`, `vgg` or `inception`
    batch_size : integer
        Batch size used for training the model
    dropout : float
        Dropout rate used for training the model
    learning_rate : float
        Learning rate used for training the model
    learning_rate_decay : float
        Learning rate decay used for training the model
    output_dir : str
        Path of the output directory, where labelled images will be stored
    (useful only if `problem=semantic_segmentation`)

    Returns
    -------
    dict
        Double predictions (between 0 and 1, acts as percentages) regarding
    each labels

    """
    # `image_paths` is first got as
    # [[image1, ..., image_i], [image_j, ..., image_n]]
    image_paths = [glob.glob(f) for f in filenames]
    # then it is flattened to get a simple list
    flattened_image_paths = sum(image_paths, [])
    images = extract_images(flattened_image_paths)
    model_input_size = images.shape[1]

    instance_args = [
        name,
        model_input_size,
        network,
        batch_size,
        dropout,
        learning_rate,
        learning_rate_decay,
    ]
    instance_name = utils.list_to_str(instance_args, "_")

    prepro_folder = utils.prepare_preprocessed_folder(
        datapath, dataset, model_input_size
    )

    if os.path.isfile(prepro_folder["training_config"]):
        train_config = utils.read_config(prepro_folder["training_config"])
        label_ids = [
            x["id"] for x in train_config["labels"] if x["is_evaluate"]
        ]
        nb_labels = len(label_ids)
    else:
        raise FileNotFoundError(
            "There is no training data with the given parameters. "
            "Please generate a valid dataset before calling the program."
        )

    output_folder = utils.prepare_output_folder(datapath, dataset, model_input_size, problem)
    instance_path = os.path.join(output_folder, output_folder["best-instance"])
    dropout, network = utils.recover_instance(instance_path)
    model = init_model(
        problem,
        instance_name,
        model_input_size,
        nb_labels,
        dropout,
        network,
    )
    if os.path.isfile(output_folder["best-model"]):
        model.load_weights(output_folder["best-model"])
        logger.info(
            "Model weights have been recovered from %s",
            output_folder["best-model"],
        )
    else:
        logger.info(
            "No available trained model for this image size with optimized hyperparameters. "
            "The inference will be done on an untrained model"
        )

    y_raw_pred = model.predict(images, batch_size=2, verbose=1)

    result = {}
    if problem == "featdet":
        label_info = [
            (i["category"], utils.GetHTMLColor(i["color"]))
            for i in train_config["labels"]
        ]
        for filename, prediction in zip(flattened_image_paths, y_raw_pred):
            result[filename] = [
                (i[0], 100 * round(float(j), 2), i[1])
                for i, j in zip(label_info, prediction)
            ]
        return result
    elif problem == "semseg":
        os.makedirs(output_dir, exist_ok=True)
        predicted_labels = np.argmax(y_raw_pred, axis=3)
        encountered_labels = np.unique(predicted_labels)
        meaningful_labels = [
            x
            for i, x in enumerate(train_config["labels"])
            if i in encountered_labels
        ]
        labelled_images = np.zeros(
            shape=np.append(predicted_labels.shape, 3), dtype=np.int8
        )
        for i in range(nb_labels):
            labelled_images[predicted_labels == i] = train_config["labels"][i][
                "color"
            ]
        for predicted_labels, filename in zip(
            labelled_images, flattened_image_paths
        ):
            predicted_image = Image.fromarray(predicted_labels, "RGB")
            filename = filename.replace(".jpg", ".png")
            predicted_image_path = os.path.join(
                output_dir, os.path.basename(filename)
            )
            predicted_image.save(predicted_image_path)
            result[filename] = os.path.basename(filename)
        return {
            "labels": summarize_config(meaningful_labels),
            "label_images": result,
        }
    else:
        raise ValueError(
                "Unknown model argument. Please use 'featdet' or 'semseg'."
        )


def summarize_config(config):
    """Extract and reshape dataset configuration information in a HTML-printing
    context

    Parameters
    ----------
    config : dict
        Dataset label configuration

    Returns
    -------
    dict
        Simplified dataset configuration for HTML-printing purpose
    """
    return [(c["category"], utils.GetHTMLColor(c["color"])) for c in config]


def extract_images(image_paths):
    """Convert a list of image filenames into a numpy array that contains the
    image data

    Parameters
    ----------
    image_paths : str
        Name of the image files onto the file system

    Returns
    -------
    np.array
        Data that is contained into the image
    """
    x_test = []
    for image_path in image_paths:
        image = Image.open(image_path)
        if image.size[0] != image.size[1]:
            raise ValueError("One of the parsed images has non-squared dimensions.")
        x_test.append(np.array(image))
    return np.array(x_test)


def main(args):
    y_raw_pred = predict(
        args.image_paths,
        args.dataset,
        args.model,
        args.datapath,
        args.name,
        args.network,
        args.batch_size,
        args.dropout,
        args.learning_rate,
        args.learning_rate_decay,
    )

    logger.info(y_raw_pred)



"""Design a feature detection model with Keras API
"""

import daiquiri
import keras as K
from keras.applications import VGG16, inception_v3, resnet50

from deeposlandia.network import ConvolutionalNeuralNetwork


logger = daiquiri.getLogger(__name__)


class FeatureDetectionNetwork(ConvolutionalNeuralNetwork):
    """Class that encapsulates feature detection network creation

    Inherits from `ConvolutionalNeuralNetwork`

    Attributes
    ----------
    network_name : str
        Name of the network
    image_size : integer
        Input image size (height and width are equal)
    nb_channels : integer
        Number of input image channels (1 for greyscaled images, 3 for RGB
    images)
    nb_labels : integer
        Number of classes in the dataset glossary
    X : tensor
        (batch_size, image_size, image_size, nb_channels)-shaped input tensor;
    input image data
    Y : tensor
        (None, nb_classes)-shaped output tensor, built as the output of the
    last network layer
    """

    def __init__(
        self,
        network_name="mapillary",
        image_size=512,
        nb_channels=3,
        nb_labels=65,
        dropout=1.0,
        architecture="simple",
    ):
        super().__init__(
            network_name, image_size, nb_channels, nb_labels, dropout
        )
        if architecture == "vgg":
            self.Y = self.vgg16()
        elif architecture == "inception":
            self.Y = self.inception()
        elif architecture == "resnet":
            self.Y = self.resnet()
        elif architecture == "simple":
            self.Y = self.simple()
        else:
            logger.error(
                (
                    "Unknown network architecture. Please use "
                    "'simple', 'vgg', 'inception' or 'resnet'."
                )
            )
            raise ValueError("Unknown network architecture.")

    def output_layer(self, x, depth):
        """Build an output layer to a neural network, as a dense layer with
        sigmoid activation function as the point is to detect multiple labels
        on a single image

        Parameters
        ----------
        x : tensor
            Previous layer within the neural network (last hidden layer)
        depth : integer
            Dimension of the previous neural network layer

        Returns
        -------
        tensor
            2D output layer
        """
        y = K.layers.Dense(depth, name="output_fc")(x)
        y = K.layers.Activation(
            activation="sigmoid", name="output_activation"
        )(y)
        return y

    def simple(self):
        """Build a simple default convolutional neural network composed of 3
        convolution-maxpool blocks and 1 fully-connected layer

        Returns
        -------
        tensor
            (batch_size, nb_labels)-shaped output predictions, that have to be
        compared with ground-truth values
        """
        layer = self.convolution(
            self.X, nb_filters=16, kernel_size=7, block_name="conv1"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool1")
        layer = self.convolution(
            layer, nb_filters=32, kernel_size=5, block_name="conv2"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool2")
        layer = self.convolution(
            layer, nb_filters=64, kernel_size=3, block_name="conv3"
        )
        layer = self.maxpool(layer, pool_size=2, strides=2, block_name="pool3")
        layer = self.flatten(layer, block_name="flatten1")
        layer = self.dense(layer, depth=512, block_name="fc1")
        return self.output_layer(layer, depth=self.nb_labels)

    def vgg16(self):
        """Build the structure of a convolutional neural network from input
        image data to the last hidden layer on the model of a similar manner
        than VGG-net

        See: Simonyan & Zisserman, Very Deep Convolutional Networks for
        Large-Scale Image Recognition, arXiv technical report, 2014

        Returns
        -------
        tensor
            (batch_size, nb_labels)-shaped output predictions, that have to be
        compared with ground-truth values
        """
        vgg16_model = VGG16(input_tensor=self.X, include_top=False)
        y = self.flatten(vgg16_model.output, block_name="flatten")
        y = self.dense(y, 1024, block_name="fc1")
        y = self.dense(y, 1024, block_name="fc2")
        return self.output_layer(y, depth=self.nb_labels)

    def resnet(self):
        """Build the structure of a convolutional neural network from input
        image data to the last hidden layer on a similar manner than ResNet

        See: He, Zhang, Ren, Sun. Deep Residual Learning for Image
        Recognition. ArXiv technical report, 2015.

        Returns
        -------
        tensor
            (batch_size, nb_labels)-shaped output predictions, that have to be
        compared with ground-truth values
        """
        resnet_model = resnet50.ResNet50(
            include_top=False, input_tensor=self.X
        )
        y = self.flatten(resnet_model.output)
        return self.output_layer(y, depth=self.nb_labels)

    def inception(self):
        """Build the structure of a convolutional neural network from input
        image data to the last hidden layer on the model of a similar manner
        than Inception-V4

        See: Szegedy, Vanhoucke, Ioffe, Shlens. Rethinking the Inception
        Architecture for Computer Vision. ArXiv technical report, 2015.

        Returns
        -------
        tensor
            (batch_size, nb_labels)-shaped output predictions, that have to be
        compared with ground-truth values

        """
        inception_model = inception_v3.InceptionV3(
            input_tensor=self.X, include_top=False
        )
        y = K.layers.GlobalAveragePooling2D()(inception_model.output)
        return self.output_layer(y, depth=self.nb_labels)



from deeposlandia.webapp.main import app

PREFIX = "/deeposlandia"
app.config["APPLICATION_ROOT"] = PREFIX
app.config["SECRET_KEY"] = "D33p0s14ndIA!"


class ReverseProxied:
    """Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    inspired from http://flask.pocoo.org/snippets/35/

    nginx example
    location /myprefix {
        proxy_pass http://192.168.0.1:5001;
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Scheme $scheme;
        proxy_set_header X-Script-Name /myprefix;
        }

    :param app: the WSGI application
    """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME", PREFIX)
        if script_name:
            environ["SCRIPT_NAME"] = PREFIX
            path_info = environ["PATH_INFO"]
            if path_info.startswith(script_name):
                environ["PATH_INFO"] = path_info[len(script_name):]

        scheme = environ.get("HTTP_X_SCHEME", "")
        if scheme:
            environ["wsgi.url_scheme"] = scheme
        return self.app(environ, start_response)


app.wsgi_app = ReverseProxied(app.wsgi_app)



"""Define data generator to feed Keras models
"""

import numpy as np

from keras.preprocessing.image import ImageDataGenerator

from deeposlandia import AVAILABLE_MODELS
from deeposlandia.datasets import AVAILABLE_DATASETS


def recover_label_id(img, label_config):
    """Recover label ids starting from a typical RGB-colored image modeled as a
    (width, height, 3)-shaped array

    Parameters
    ----------
    img : np.array
        Image data of shape (width, heigth, 3)
    label_config : dict
        Labels contained into the dataset

    Returns
    -------
    np.array
        Array of pixel labels of shape (width, heigth)
    """
    reshaped_img = img.reshape([-1, 3])
    label_img = np.full(
        shape=reshaped_img.shape[0], fill_value=-1, dtype=reshaped_img.dtype
    )
    for label in label_config:
        label_img[np.all(label["color"] == reshaped_img, axis=1)] = label["id"]
    return label_img.reshape(img.shape[:3])


def feature_detection_labelling(img, label_config):
    """One-hot encoding for feature detection problem

    Parameters
    ----------
    img : numpy.array
        Batched input image data of size (batch_size, image_size, image_size,
    1)
    label_config : dict
        Labels contained into the dataset

    Returns
    -------
    numpy.array
        Label encoding, array of shape (batch_size, nb_labels)
    """
    if not (img.shape[1] == img.shape[2] and len(img.shape) == 4):
        raise ValueError(
            (
                "Wrong image shape. Please provide batched RGB-"
                "images with equal width and height dimensions."
            )
        )
    label_ids = [item["id"] for item in label_config if item["is_evaluate"]]
    if not all(
        isinstance(item, (int, np.uint8, np.uint32, np.uint64))
        for item in label_ids
    ):
        raise ValueError(
            ("List of label IDs must contains " "integers: {}").format(
                label_ids
            )
        )
    img = recover_label_id(img, label_config)
    flattened_images = img.reshape(img.shape[0], -1).astype(np.uint8)
    one_hot_encoding = np.equal.outer(flattened_images, label_ids)
    return one_hot_encoding.any(axis=1)


def semantic_segmentation_labelling(img, label_config):
    """One-hot encoder for semantic segmentation problem

    Parameters
    ----------
    img : ndarray
        Batched input image data of size (batch_size, image_size, image_size,
    1)
    label_config : dict
        Label contained into the dataset

    Returns
    -------
    ndarray
        Label encoding, array of shape (batch_size, image_size, image_size,
    nb_labels), occurrence of the ith label for each pixel
    """
    if not (img.shape[1] == img.shape[2] and len(img.shape) == 4):
        raise ValueError(
            (
                "Wrong image shape. Please provide batched RGB-"
                "images with equal width and height dimensions."
            )
        )
    label_ids = [item["id"] for item in label_config if item["is_evaluate"]]
    if not all(
        isinstance(item, (int, np.uint8, np.uint32, np.uint64))
        for item in label_ids
    ):
        raise ValueError(
            ("List of label IDs must contains " "integers: {}").format(
                label_ids
            )
        )
    img = recover_label_id(img, label_config)
    img = img.squeeze().astype(np.uint8)
    one_hot_encoding = np.equal.outer(img, label_ids)
    return one_hot_encoding


def feed_generator(datapath, gen_type, image_size, batch_size, seed=None):
    """Build a couple of generator fed by image and label repository,
    respectively

    The input image are stored as RGB-images, whilst labelled image are
    grayscaled-images. The Keras generator takes this difference into account
    through its `color_mode` parameter, that depends on `gen_type`

    Parameters
    ----------
    datapath : str
        Path to image repository
    gen_type : str
        Generator type, either `images` or `labels`
    image_size : integer
        Number of width (resp. height) pixels
    batch_size : integer
        Number of images in each training batch
    they are stored as RGB pixels
    seed : integer
        Random number generation for data shuffling and transformations

    Returns
    -------
    generator
        Input image generator
    """
    generator = ImageDataGenerator()
    return generator.flow_from_directory(
        datapath,
        classes=[gen_type],
        target_size=(image_size, image_size),
        batch_size=batch_size,
        class_mode=None,
        color_mode="rgb",
        seed=seed,
    )


def create_generator(
    dataset,
    model,
    datapath,
    image_size,
    batch_size,
    label_config,
    inference=False,
    seed=None,
):
    """Create a Keras data Generator starting from images contained in
    `datapath` repository to address `model`

    Parameters
    ----------
    dataset : str
        Name of the dataset (*e.g.* `shapes`, `mapillary`, `aerial` or
    `tanzania`)
    model : str
        Research problem that is addressed (either `feature_detection` or
    `semantic_segmentation`)
    datapath : str
        Path to image repository
    image_size : integer
        Number of width (resp. height) pixels
    batch_size : integer
        Number of images in each training batch
    label_config : dict
        Dataset valid label description
    inference : boolean
        If True, generates only image data (labels are not considered during
    inference)
    seed : integer
        Random number generation for data shuffling and transformations

    Returns
    -------
    generator
        Generator of tuples (images, labels), for each input data batch

    """
    if dataset not in AVAILABLE_DATASETS:
        raise ValueError("Wrong dataset name {}".format(dataset))
    image_generator = feed_generator(
        datapath, "images", image_size, batch_size, seed
    )
    if inference:
        return image_generator
    label_generator = feed_generator(
        datapath, "labels", image_size, batch_size, seed
    )
    if model == "featdet":
        label_generator = (
            feature_detection_labelling(x, label_config)
            for x in label_generator
        )
    elif model == "semseg":
        label_generator = (
            semantic_segmentation_labelling(x, label_config)
            for x in label_generator
        )
    else:
        raise ValueError(
            "Wrong model name {} (choose amongst {})".format(
                model, AVAILABLE_MODELS
            )
        )
    return zip(image_generator, label_generator)



"""Mapillary dataset module

Model Mapillary street-scene images for convolutional neural network
applications. Data is downloadable at https://www.mapillary.com/dataset/vistas.

"""

from multiprocessing import Pool
import os

import daiquiri
import numpy as np
from PIL import Image

from deeposlandia.datasets import Dataset
from deeposlandia import utils


logger = daiquiri.getLogger(__name__)


class MapillaryDataset(Dataset):
    """Dataset structure that gathers all information related to the Mapillary
    images

    Attributes
    ----------
    image_size : int
        Size of considered images (height=width), raw images will be resized
    during the preprocessing
    glossary_filename : str
        Name of the Mapillary input glossary, that contains every information
    about Mapillary labels

    """

    def __init__(self, image_size, glossary_filename):
        """ Class constructor ; instanciates a MapillaryDataset as a standard
        Dataset which is completed by a glossary file that describes the
        dataset labels
        """
        super().__init__(image_size)
        self.build_glossary(glossary_filename)

    def build_glossary(self, config_filename):
        """Read the Mapillary glossary stored as a json file at the data
        repository root

        Parameters
        ----------
        config_filename : str
            String designing the relative path of the dataset glossary
        (based on Mapillary dataset)
        """
        glossary = utils.read_config(config_filename)
        if "labels" not in glossary:
            logger.error("There is no 'label' key in the provided glossary.")
            return None
        for lab_id, label in enumerate(glossary["labels"]):
            self.add_label(
                lab_id,
                label["name"],
                label["color"],
                label["evaluate"],
                label["family"],
                label["contains_id"],
                label["contains"]
            )

    def group_image_label(self, image):
        """Group the labels

        If the label ids 4, 5 and 6 belong to the same group, they will be
        turned into the label id 4.

        Parameters
        ----------
        image : PIL.Image

        Returns
        -------
        PIL.Image
        """
        # turn all label ids into the lowest digits/label id
        # according to its "group" (manually built)
        a = np.array(image)
        for root_id, label in enumerate(self.label_info):
            for label_id in label.get("aggregate"):
                mask = a == label_id
                a[mask] = root_id
        return Image.fromarray(a, mode=image.mode)

    def _preprocess(
        self, image_filename, output_dir, labelling=True
    ):
        """Resize/crop then save the training & label images

        Parameters
        ----------
        datadir : str
        image_filaname : str
        labelling : boolean

        Returns
        -------
        dict
            Key/values with the filenames and label ids
        """
        # open original images
        img_in = Image.open(image_filename)

        # resize images
        # (self.image_size*larger_size or larger_size*self.image_size)
        img_in = utils.resize_image(img_in, self.image_size)

        # crop images to get self.image_size*self.image_size dimensions
        crop_pix = np.random.randint(0, 1 + max(img_in.size) - self.image_size)
        final_img_in = utils.mono_crop_image(img_in, crop_pix)

        # save final image
        new_in_filename = os.path.join(
            output_dir, "images", os.path.basename(image_filename)
        )
        final_img_in.save(new_in_filename)

        # label_filename vs label image
        if labelling:
            label_filename = image_filename.replace("images/", "labels/")
            label_filename = label_filename.replace(".jpg", ".png")
            img_out = Image.open(label_filename)
            img_out = utils.resize_image(img_out, self.image_size)
            img_out = utils.mono_crop_image(img_out, crop_pix)
            # aggregate some labels
            img_out = self.group_image_label(img_out)

            labels = utils.build_labels(
                img_out, self.label_ids, dataset="mapillary"
            )
            new_out_filename = os.path.join(
                output_dir, "labels", os.path.basename(label_filename)
            )
            label_out = np.array(img_out)
            final_img_out = utils.build_image_from_config(
                label_out, self.label_info
            )
            final_img_out.save(new_out_filename)
        else:
            new_out_filename = None
            labels = {i: 0 for i in range(self.get_nb_labels())}

        return {
            "raw_filename": image_filename,
            "image_filename": new_in_filename,
            "label_filename": new_out_filename,
            "labels": labels,
        }

    def populate(
        self,
        output_dir,
        input_dir,
        nb_images=None,
        nb_tiles_per_image=None,
        labelling=True,
        nb_processes=1,
    ):
        """ Populate the dataset with images contained into `datadir` directory

        Parameters
        ----------
        output_dir : str
            Path of the directory where the preprocessed image must be saved
        input_dir : str
            Path of the directory that contains input images
        nb_images : integer
            Number of images to be considered in the dataset; if None, consider the whole
        repository
        nb_tiles_per_image : integer
            Number of tiles that must be picked into the raw image (useless there, added
        for consistency)
        labelling: boolean
            If True labels are recovered from dataset, otherwise dummy label are generated
        nb_processes : int
            Number of processes on which to run the preprocessing
        """
        if nb_tiles_per_image is not None:
            logger.warning("The ``nb_tiles_per_image`` parameter is useless, it will be ignored.")
        image_list = os.listdir(os.path.join(input_dir, "images"))[:nb_images]
        image_list_longname = [
            os.path.join(input_dir, "images", l) for l in image_list
        ]
        if nb_processes == 1:
            for x in image_list_longname:
                self.image_info.append(
                    self._preprocess(x, output_dir, labelling)
                )
        else:
            with Pool(processes=nb_processes) as p:
                self.image_info = p.starmap(
                    self._preprocess,
                    [
                        (x, output_dir, labelling)
                        for x in image_list_longname
                    ],
                )



"""Main method to train neural networks with Keras API
"""

import argparse
from datetime import datetime
import itertools
import json
import os

import daiquiri
import numpy as np

from keras import backend, callbacks
from keras.models import Model
from keras.optimizers import Adam

from deeposlandia import generator, utils
from deeposlandia import AVAILABLE_MODELS
from deeposlandia.datasets import AVAILABLE_DATASETS
from deeposlandia.feature_detection import FeatureDetectionNetwork
from deeposlandia.semantic_segmentation import SemanticSegmentationNetwork
from deeposlandia.metrics import iou, dice_coef

SEED = int(datetime.now().timestamp())


logger = daiquiri.getLogger(__name__)


def get_data(folders, dataset, model, image_size, batch_size):
    """On the file system, recover `dataset` that can solve `model` problem

    Parameters
    ----------
    folders : dict
        Dictionary of useful folders that indicates paths to data
    dataset : str
        Name of the used dataset (*e.g.* `shapes` or `mapillary`)
    model : str
        Name of the addressed research problem (*e.g.* `feature_detection` or
    `semantic_segmentation`)
    image_size : int
        Size of the images, in pixel (height=width)
    batch_size : int
        Number of images in each batch

    Returns
    -------
    tuple
        Number of labels in the dataset, as well as training and validation
    data generators

    """
    # Data gathering
    if os.path.isfile(folders["training_config"]):
        train_config = utils.read_config(folders["training_config"])
        label_ids = [
            x["id"] for x in train_config["labels"] if x["is_evaluate"]
        ]
        train_generator = generator.create_generator(
            dataset,
            model,
            folders["training"],
            image_size,
            batch_size,
            train_config["labels"],
            seed=SEED,
        )
    else:
        raise FileNotFoundError(
            "There is no training data with the given parameters. Please "
            "generate a valid dataset before calling the training program."
        )
    if os.path.isfile(folders["validation_config"]):
        validation_generator = generator.create_generator(
            dataset,
            model,
            folders["validation"],
            image_size,
            batch_size,
            train_config["labels"],
            seed=SEED,
        )
    else:
        raise FileNotFoundError(
            "There is no validation data with the given parameters. Please "
            "generate a valid dataset before calling the training program."
        )
    nb_labels = len(label_ids)
    return nb_labels, train_generator, validation_generator


def run_model(
    train_generator,
    validation_generator,
    dl_model,
    output_folder,
    instance_name,
    image_size,
    nb_labels,
    nb_epochs,
    nb_training_image,
    nb_validation_image,
    batch_size,
    dropout,
    network,
    learning_rate,
    learning_rate_decay,
):
    """Run deep learning `dl_model` starting from training and validation data
    generators, depending on a range of hyperparameters

    Parameters
    ----------
    train_generator : generator
        Training data generator
    validation_generator : generator
        Validation data generator
    dl_model : str
        Name of the addressed research problem (*e.g.* `feature_detection` or
    `semantic_segmentation`)
    output_folder : str
        Name of the folder where the trained model will be stored on the file
    system
    instance_name : str
        Name of the instance
    image_size : int
        Size of images, in pixel (height=width)
    nb_labels : int
        Number of labels into the dataset
    nb_epochs : int
        Number of epochs during which models will be trained
    nb_training_image : int
        Number of images into the training dataset
    nb_validation_image : int
        Number of images into the validation dataset
    batch_size : int
        Number of images into each batch
    dropout : float
        Probability of keeping a neuron during dropout phase
    network : str
        Neural network architecture (*e.g.* `simple`, `vgg`, `inception`)
    learning_rate : float
        Starting learning rate
    learning_rate_decay : float
        Learning rate decay

    Returns
    -------
    dict
        Dictionary that summarizes the instance and the corresponding model
    performance (measured by validation accuracy)
    """
    if dl_model == "featdet":
        net = FeatureDetectionNetwork(
            network_name=instance_name,
            image_size=image_size,
            nb_channels=3,
            nb_labels=nb_labels,
            architecture=network,
        )
        loss_function = "binary_crossentropy"
    elif dl_model == "semseg":
        net = SemanticSegmentationNetwork(
            network_name=instance_name,
            image_size=image_size,
            nb_channels=3,
            nb_labels=nb_labels,
            architecture=network,
        )
        loss_function = "categorical_crossentropy"
    else:
        raise ValueError(
            f"Unrecognized model: {dl_model}. Please choose amongst {AVAILABLE_MODELS}"
        )
    model = Model(net.X, net.Y)
    opt = Adam(lr=learning_rate, decay=learning_rate_decay)
    metrics = ["acc", iou, dice_coef]
    model.compile(loss=loss_function, optimizer=opt, metrics=metrics)

    # Model training
    steps = max(nb_training_image // batch_size, 1)
    val_steps = max(nb_validation_image // batch_size, 1)

    checkpoint_files = [
        item
        for item in os.listdir(output_folder)
        if "checkpoint-epoch" in item
    ]
    if len(checkpoint_files) > 0:
        model_checkpoint = max(checkpoint_files)
        trained_model_epoch = int(model_checkpoint[-5:-3])
        checkpoint_complete_path = os.path.join(
            output_folder, model_checkpoint
        )
        model.load_weights(checkpoint_complete_path)
        logger.info(
            "Model weights have been recovered from %s",
            checkpoint_complete_path,
        )
    else:
        logger.info(
            (
                "No available checkpoint for this configuration. "
                "The model will be trained from scratch."
            )
        )
        trained_model_epoch = 0

    checkpoint_filename = os.path.join(
        output_folder, "checkpoint-epoch-{epoch:03d}.h5"
    )
    checkpoint = callbacks.ModelCheckpoint(
        checkpoint_filename,
        monitor="val_loss",
        verbose=0,
        save_best_only=True,
        save_weights_only=False,
        mode="auto",
        period=1,
    )
    terminate_on_nan = callbacks.TerminateOnNaN()
    earlystop = callbacks.EarlyStopping(
        monitor="val_loss", patience=10, verbose=1, mode="max"
    )
    csv_logger = callbacks.CSVLogger(
        os.path.join(output_folder, "training_metrics.csv"), append=True
    )

    hist = model.fit_generator(
        train_generator,
        epochs=nb_epochs,
        initial_epoch=trained_model_epoch,
        steps_per_epoch=steps,
        validation_data=validation_generator,
        validation_steps=val_steps,
        callbacks=[checkpoint, earlystop, terminate_on_nan, csv_logger],
    )
    ref_metric = max(hist.history.get("val_acc", [np.nan]))
    return {
        "model": model,
        "val_acc": ref_metric,
        "batch_size": batch_size,
        "network": network,
        "dropout": dropout,
        "learning_rate": learning_rate,
        "learning_rate_decay": learning_rate_decay,
    }


def main(args):
    # Grid search
    model_output = []
    output_folder = utils.prepare_output_folder(
        args.datapath, args.dataset, args.image_size, args.model
    )
    for batch_size in args.batch_size:
        logger.info("Generating data with batch of %s images...", batch_size)
        # Data generator building
        prepro_folder = utils.prepare_preprocessed_folder(
            args.datapath, args.dataset, args.image_size
        )
        nb_labels, train_gen, valid_gen = get_data(
            prepro_folder,
            args.dataset,
            args.model,
            args.image_size,
            batch_size,
        )
        for parameters in itertools.product(
            args.dropout,
            args.network,
            args.learning_rate,
            args.learning_rate_decay,
        ):
            logger.info("Instance: %s", utils.list_to_str(parameters))
            # Data path and repository management
            dropout, network, learning_rate, learning_rate_decay = parameters
            instance_args = [
                args.name,
                args.image_size,
                network,
                batch_size,
                dropout,
                learning_rate,
                learning_rate_decay,
            ]
            instance_name = utils.list_to_str(instance_args, "_")
            instance_folder = os.path.join(output_folder["checkpoints"], instance_name)
            # Model running
            model_output.append(
                run_model(
                    train_gen,
                    valid_gen,
                    args.model,
                    instance_folder,
                    instance_name,
                    args.image_size,
                    nb_labels,
                    args.nb_epochs,
                    args.nb_training_image,
                    args.nb_validation_image,
                    batch_size,
                    *parameters
                )
            )
            logger.info("Instance result: %s", model_output[-1])

    # Recover best instance starting from validation accuracy
    best_instance = max(model_output, key=lambda x: x["val_acc"])

    # Save best model
    best_instance["model"].save(output_folder["best-model"])
    with open(output_folder["best-instance"], "w") as fobj:
        json.dump(
            {
                key: best_instance[key]
                for key in best_instance
                if key != "model"
            },
            fobj,
        )

    backend.clear_session()



"""Deeposlandia package
"""

from configparser import ConfigParser
import logging
import os

import daiquiri

__version__ = "0.6.post3"


# Do not log Tensorflow messages
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

# Configure the logger
daiquiri.setup(
    level=logging.INFO,
    outputs=(
        daiquiri.output.Stream(
            formatter=daiquiri.formatter.ColorFormatter(
                fmt=(
                    "%(asctime)s :: %(levelname)s :: %(module)s :: "
                    "%(funcName)s : %(color)s%(message)s%(color_stop)s"
                )
            )
        ),
    ),
)
logger = daiquiri.getLogger("root")

# Deeposlandia supports feature detection (featdet) and semantic segmentation (semseg)
AVAILABLE_MODELS = ("featdet", "semseg")

# Configuration file handling
_DEEPOSL_CONFIG = os.getenv("DEEPOSL_CONFIG")
_DEEPOSL_CONFIG = (
    _DEEPOSL_CONFIG if _DEEPOSL_CONFIG is not None else "config.ini"
)
config = ConfigParser()
if os.path.isfile(_DEEPOSL_CONFIG):
    config.read(_DEEPOSL_CONFIG)
else:
    raise FileNotFoundError("No file config.ini!")



import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def find_version(*file_paths):
    with open(os.path.join(here, *file_paths), "r") as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split("=")[1].strip(" '\"")
    raise RuntimeError(
        ("Unable to find version string. " "Should be __init__.py.")
    )


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    "tensorflow==2.0.1",
    "opencv-python<=4.2.0.34",
    "numpy<=1.16.2",
    "pillow<=7.1.1",
    "keras<=2.3.1",
    "daiquiri<=2.1.1",
    "Flask<=1.0.2",
    "seaborn<=0.8.1",
    "shapely<=1.6.0",
    "geopandas<=0.7.0",
    "rtree<=0.9.4",
]

setup(
    name="deeposlandia",
    keywords=[
        "deep learning",
        "convolutional neural networks",
        "image",
        "Keras",
    ],
    version=find_version("deeposlandia", "__init__.py"),
    description=(
        "Automatic detection and semantic image segmentation "
        "with deep learning"
        ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Oslandia",
    author_email="info@oslandia.com",
    maintainer="Oslandia",
    maintainer_email="",
    url="https://github.com/Oslandia/deeposlandia",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3",
    install_requires=install_requires,
    packages=find_packages(),
    entry_points = {
        "console_scripts": ["deepo=deeposlandia.__main__:main"],
    }
)



"""Aerial dataset module

Model Aerial images for convolutional neural network applications. Data is
downloadable at https://project.inria.fr/aerialimagelabeling/files/.

"""

from multiprocessing import Pool
import os

import daiquiri
import numpy as np
from osgeo import gdal
from PIL import Image

from deeposlandia.datasets import GeoreferencedDataset
from deeposlandia import geometries, utils


logger = daiquiri.getLogger(__name__)


class AerialDataset(GeoreferencedDataset):
    """Dataset structure inspired from AerialImageDataset, a dataset released
    by Inria

    The dataset is freely available at:
    https://project.inria.fr/aerialimagelabeling/files/

    It is composed of 180 training images and 180 testing images of size
    5000*5000. There are ground-truth labels only for the former set.

    Attributes
    ----------
    tile_size : int
        Size of the tiles into which each raw images is decomposed during
    dataset population (height=width)
    """

    def __init__(self, tile_size):
        """ Class constructor ; instanciates a AerialDataset as a standard
        Dataset which is completed by a glossary file that describes the
        dataset labels and images
        """
        # self.tile_size = tile_size
        # img_size = utils.get_image_size_from_tile(self.tile_size)
        super().__init__(tile_size)
        self.add_label(
            label_id=0, label_name="background", color=0, is_evaluate=True
        )
        self.add_label(
            label_id=1, label_name="building", color=255, is_evaluate=True
        )


    def _preprocess_for_training(self, image_filename, output_dir, nb_images):
        """Resize/crop then save the training & label images

        Parameters
        ----------
        image_filename : str
            Full path towards the image on the disk
        output_dir : str
            Output path where preprocessed image must be saved

        Returns
        -------
        dict
            Key/values with the filenames and label ids
        """
        raster = gdal.Open(image_filename)
        raw_img_width = raster.RasterXSize
        raw_img_height = raster.RasterYSize
        image_data = raster.ReadAsArray()
        image_data = np.swapaxes(image_data, 0, 2)
        result_dicts = []
        logger.info(
            "Image filename: %s, size: (%s, %s)",
            image_filename.split("/")[-1], raw_img_width, raw_img_height
        )

        label_filename = image_filename.replace("images", "labels")
        label_raster = gdal.Open(label_filename)
        labels = label_raster.ReadAsArray()
        labels = np.swapaxes(labels, 0, 1)

        nb_attempts = 0
        image_counter = 0
        empty_image_counter = 0
        while image_counter < nb_images and nb_attempts < 2 * nb_images:
            # randomly pick an image
            x = np.random.randint(0, raw_img_width - self.image_size)
            y = np.random.randint(0, raw_img_height - self.image_size)

            tile_data = image_data[
                x:(x + self.image_size), y:(y + self.image_size)
            ]
            tile_image = Image.fromarray(tile_data)
            mask = labels[
                x:(x + self.image_size), y:(y + self.image_size)
            ]
            label_dict = utils.build_labels(
                mask, range(self.get_nb_labels()), "aerial"
            )
            labelled_image = Image.fromarray(mask)
            if np.unique(mask).shape[0] > 1:
                tiled_results = self._serialize(
                    tile_image,
                    labelled_image,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "nw",
                )
                if tiled_results:
                    result_dicts.append(tiled_results)
                image_counter += 1
                tile_image_ne = tile_image.transpose(Image.FLIP_LEFT_RIGHT)
                labelled_image_ne = labelled_image.transpose(
                    Image.FLIP_LEFT_RIGHT
                )
                tiled_results_ne = self._serialize(
                    tile_image_ne,
                    labelled_image_ne,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "ne",
                )
                if tiled_results_ne:
                    result_dicts.append(tiled_results_ne)
                image_counter += 1
                tile_image_sw = tile_image.transpose(Image.FLIP_TOP_BOTTOM)
                labelled_image_sw = labelled_image.transpose(
                    Image.FLIP_TOP_BOTTOM
                )
                tiled_results_sw = self._serialize(
                    tile_image_sw,
                    labelled_image_sw,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "sw",
                )
                if tiled_results_sw:
                    result_dicts.append(tiled_results_sw)
                image_counter += 1
                tile_image_se = tile_image_sw.transpose(Image.FLIP_LEFT_RIGHT)
                labelled_image_se = labelled_image_sw.transpose(
                    Image.FLIP_LEFT_RIGHT
                )
                tiled_results_se = self._serialize(
                    tile_image_se,
                    labelled_image_se,
                    label_dict,
                    image_filename,
                    output_dir,
                    x,
                    y,
                    "se",
                )
                if tiled_results_se:
                    result_dicts.append(tiled_results_se)
                image_counter += 1
                del tile_image_se, tile_image_sw, tile_image_ne
                del labelled_image_se, labelled_image_sw, labelled_image_ne
            else:
                if empty_image_counter < 0.1 * nb_images:
                    tiled_results = self._serialize(
                        tile_image,
                        labelled_image,
                        label_dict,
                        image_filename,
                        output_dir,
                        x,
                        y,
                        "nw",
                    )
                    if tiled_results:
                        result_dicts.append(tiled_results)
                    image_counter += 1
                    empty_image_counter += 1
            nb_attempts += 1
        del raster
        logger.info(
            "Generate %s images after %s attempts.", image_counter, nb_attempts
        )
        return result_dicts



"""Main method to generate new datasets

Example of program call:

* generate 64*64 pixel images from Shapes dataset, 10000 images in the training
  set, 100 in the validation set, 1000 in the testing set::

    python deeposlandia/datagen.py -D shapes -s 64 -t 10000 -v 100 -T 1000

"""

import argparse
import os

import daiquiri
import pandas as pd

from deeposlandia import config, utils
from deeposlandia.datasets import AVAILABLE_DATASETS, GEOGRAPHIC_DATASETS
from deeposlandia.datasets.mapillary import MapillaryDataset
from deeposlandia.datasets.aerial import AerialDataset
from deeposlandia.datasets.shapes import ShapeDataset
from deeposlandia.datasets.tanzania import TanzaniaDataset


logger = daiquiri.getLogger(__name__)


def main(args):
    # Data path and repository management
    input_folder = utils.prepare_input_folder(args.datapath, args.dataset)
    prepro_folder = utils.prepare_preprocessed_folder(
        args.datapath, args.dataset, args.image_size
    )
    if (
            args.dataset in GEOGRAPHIC_DATASETS
            and (args.nb_training_image > 0 or args.nb_validation_image > 0)
            and args.nb_tiles_per_image is None
    ):
        raise ValueError(
            "The amount of tiles per image must be specified for "
            f"the {args.dataset} dataset, if training and/or validation images "
            "are required. See 'deepo datagen -h' for more details."
        )

    # Dataset creation
    if args.dataset == "mapillary":
        config_path = os.path.join(input_folder, "config_aggregate.json")
        train_dataset = MapillaryDataset(args.image_size, config_path)
        validation_dataset = MapillaryDataset(args.image_size, config_path)
        test_dataset = MapillaryDataset(args.image_size, config_path)
    elif args.dataset == "shapes":
        train_dataset = ShapeDataset(args.image_size)
        validation_dataset = ShapeDataset(args.image_size)
        test_dataset = ShapeDataset(args.image_size)
        os.makedirs(
            os.path.join(prepro_folder["testing"], "labels"), exist_ok=True
        )
    elif args.dataset == "aerial":
        train_dataset = AerialDataset(args.image_size)
        validation_dataset = AerialDataset(args.image_size)
        test_dataset = AerialDataset(args.image_size)
    elif args.dataset == "tanzania":
        train_dataset = TanzaniaDataset(args.image_size)
        validation_dataset = TanzaniaDataset(args.image_size)
        test_dataset = TanzaniaDataset(args.image_size)
    else:
        raise ValueError(
            f"Unsupported dataset type. Please choose amongst {AVAILABLE_DATASETS}"
        )

    # Dataset populating/loading
    # (depends on the existence of a specification file)
    if args.nb_training_image > 0:
        if os.path.isfile(prepro_folder["training_config"]):
            train_dataset.load(
                prepro_folder["training_config"], args.nb_training_image
            )
        else:
            logger.info(
                    "No existing configuration file for this dataset. Create %s.",
                    prepro_folder["training_config"],
            )
            input_image_dir = os.path.join(input_folder, "training")
            train_dataset.populate(
                prepro_folder["training"],
                input_image_dir,
                nb_images=args.nb_training_image,
                nb_processes=int(config.get("running", "processes")),
                nb_tiles_per_image=args.nb_tiles_per_image,
            )
            train_dataset.save(prepro_folder["training_config"])

    if args.nb_validation_image > 0:
        if os.path.isfile(prepro_folder["validation_config"]):
            validation_dataset.load(
                prepro_folder["validation_config"], args.nb_validation_image
            )
        else:
            logger.info(
                    "No existing configuration file for this dataset. Create %s.",
                    prepro_folder["validation_config"],
            )
            input_image_dir = os.path.join(input_folder, "validation")
            validation_dataset.populate(
                prepro_folder["validation"],
                input_image_dir,
                nb_images=args.nb_validation_image,
                nb_processes=int(config.get("running", "processes")),
                nb_tiles_per_image=args.nb_tiles_per_image,
            )
            validation_dataset.save(prepro_folder["validation_config"])

    if args.nb_testing_image > 0:
        if os.path.isfile(prepro_folder["testing_config"]):
            test_dataset.load(
                prepro_folder["testing_config"], args.nb_testing_image
            )
        else:
            logger.info(
                    "No existing configuration file for this dataset. Create %s.",
                    prepro_folder["testing_config"],
            )
            input_image_dir = os.path.join(input_folder, "testing")
            test_dataset.populate(
                prepro_folder["testing"],
                input_image_dir,
                nb_images=args.nb_testing_image,
                labelling=False,
                nb_processes=int(config.get("running", "processes")),
            )
            test_dataset.save(prepro_folder["testing_config"])

    glossary = pd.DataFrame(train_dataset.labels)
    glossary["popularity"] = train_dataset.get_label_popularity()
    logger.info("Data glossary:\n%s", glossary)



"""Post-process predicted aerial tiles in order to rebuild complete high
resolution image.

Do inference on test images starting from a trained model and Keras API. Then
handle geographic datasets with dedicated libraries (GDAL, geopandas, shapely)
"""

import argparse
import glob
import os

import daiquiri
import geopandas as gpd
from osgeo import gdal
import numpy as np
from PIL import Image

from keras.models import Model
import keras.backend as K

from deeposlandia import utils, geometries
from deeposlandia.semantic_segmentation import SemanticSegmentationNetwork


logger = daiquiri.getLogger(__name__)


def get_image_paths(testing_folder, image_basename):
    """Returns a list with the path of every image that belongs to the
    `dataset`, preprocessed in `image_size`-pixelled images, that were
    extracted from an original image named as `image_basename`.

    Parameters
    ----------
    testing_folder : str
        Path of the testing image folder
    image_basename : str
        Original image name

    Returns
    -------
    list
        List of image full paths
    """
    image_raw_paths = os.path.join(testing_folder, "images", image_basename + "*")
    return [glob.glob(f) for f in [image_raw_paths]][0]


def extract_images(image_paths):
    """Convert a list of image filenames into a numpy array that contains the
    image data

    Parameters
    ----------
    image_paths : str
        Name of the image files onto the file system

    Returns
    -------
    np.array
        Data that is contained into the image
    """
    x_test = []
    for image_path in image_paths:
        if not image_path.endswith(".png"):
            logger.error("The filename does not refer to an png image.")
            raise ValueError()
        image = Image.open(image_path)
        if image.size[0] != image.size[1]:
            logger.error(
                "One of the parsed images has non-squared dimensions."
            )
            raise ValueError()
        x_test.append(np.array(image))
    return np.array(x_test)


def get_labels(datapath, dataset, tile_size):
    """Extract labels from the `dataset` glossary, according to the
    preprocessed version of the dataset

    Parameters
    ----------
    datapath : str
        Path of the data on the file system
    dataset : str
        Name of the dataset
    tile_size : int
        Size of preprocessed images, in pixels

    Returns
    -------
    list
        List of dictionnaries that describes the dataset labels
    """
    prepro_folder = utils.prepare_preprocessed_folder(
        datapath, dataset, tile_size,
    )
    if os.path.isfile(prepro_folder["testing_config"]):
        test_config = utils.read_config(prepro_folder["testing_config"])
    else:
        raise ValueError(
            (
                "There is no testing data with the given "
                "parameters. Please generate a valid dataset "
                "before calling the program."
            )
        )
    return [l for l in test_config["labels"] if l["is_evaluate"]]


def get_trained_model(model_filepath, image_size, nb_labels):
    """Recover model weights stored on the file system, and assign them into
    the `model` structure

    Parameters
    ----------
    model_filepath : str
        Path of the model on the file system
    image_size : int
        Image size, in pixels (height=width)
    nb_labels : int
        Number of output labels

    Returns
    -------
    keras.models.Model
        Convolutional neural network
    """
    K.clear_session()
    net = SemanticSegmentationNetwork(
        network_name="semseg_postprocessing",
        image_size=image_size,
        nb_labels=nb_labels,
        dropout=1.0,
        architecture="unet",
    )
    model = Model(net.X, net.Y)
    if os.path.isfile(model_filepath):
        model.load_weights(model_filepath)
        logger.info("Model weights have been recovered from %s" % model_filepath)
    else:
        logger.info(
            (
                "No available trained model for this image size"
                " with optimized hyperparameters. The "
                "inference will be done on an untrained model"
            )
        )
    return model


def assign_label_colors(predicted_labels, labels):
    """Transform raw Keras prediction into an exploitable numpy array that
    contains label IDs

    Parameters
    ----------
    predicted_labels : numpy.array
        Output of the prediction process, of shape (img_size, img_size)
    labels : list
        List of dictionnaries that describes the dataset labels

    Returns
    -------
    numpy.array
        Colored predicted labels, of shape (img_size, img_size, 3)
    """
    labelled_images = np.zeros(
        shape=np.append(predicted_labels.shape, 3), dtype=np.uint8
    )
    for idx, label in enumerate(labels):
        labelled_images[predicted_labels == idx] = label["color"]
    return labelled_images


def extract_coordinates_from_filenames(filenames):
    """Extract tiled image pixel coordinates starting from their filename, as
    it enrols the west and north coordinates for tile identification purpose

    Parameters
    ----------
    filenames : list
        List of tiled image filenames

    Returns
    -------
    list
        List of tiled image west and north pixel coordinates, regarding the
    full original image
    """
    basenames = [os.path.splitext(os.path.basename(f))[0] for f in filenames]
    coordinates = [b.split("_")[-2:] for b in basenames]
    return [[int(x), int(y)] for x, y in coordinates]


def fill_labelled_image(
    predictions, coordinates, tile_size, img_width, img_height=None
):
    """Fill labelled version of the original image with the labelled version of
    the corresponding tiles

    Parameters
    ----------
    predictions : numpy.array
        Labelled version of tiled images, according to semantic segmentation
    model predictions
    coordinates : list
        List of tiled image west and north pixel coordinates, regarding the
    full original image
    tile_size : int
        Tiled image size, in pixel
    img_width : int
        Original image width, in pixel
    img_height : int
        Original image height, in pixel

    Returns
    -------
    numpy.array
        Labelled version of the original image, where (i, j)-th pixel value
    corresponds to its predicted label
    """
    img_height = img_height if img_height is not None else img_width
    extended_width = img_width + tile_size - img_width % tile_size
    extended_height = img_height + tile_size - img_height % tile_size
    predicted_image = np.zeros(
        [extended_height, extended_width], dtype=np.uint8
    )
    for coords, image_data in zip(coordinates, predictions):
        x, y = coords
        predicted_image[y:(y + tile_size), x:(x + tile_size)] = image_data
    return predicted_image[:img_height, :img_width]


def build_full_labelled_image(
    images,
    coordinates,
    model,
    tile_size,
    img_width,
    img_height=None,
    batch_size=2,
):
    """Generate a full labelled version of an image (of shape (`img_width`,
        `img_height`)), knowing that it was previously splitted in tiles (of
        size `tile_size`). These tiles are stored on the file system as
        ̀image_paths`.

    This function includes a semantic segmentation model prediction step. The
    corresponding `model` is passed as another function argument, as well as
    the `batch_size` used during inference.

    Parameters
    ----------
    images : numpy.array
        Data associated to tiled images
    coordinates : list
        List of tiled image west and north pixel coordinates, regarding the
    full original image
    model : keras.models.Model
        Convolutional neural network
    tile_size : int
        Size of the tiled images, in pixel
    img_width : int
        Original image width, in pixel
    img_height : int
        Original image height, in pixel
    batch_size : int
        Number of images passed in each inference batches

    Returns
    -------
    numpy.array
        Labelled version of the original image, where (i, j)-th pixel value
    corresponds to its predicted label
    """
    img_height = img_height if img_height is not None else img_width
    y_raw_preds = model.predict(images, batch_size=batch_size, verbose=1)
    predicted_labels = np.argmax(y_raw_preds, axis=3)
    full_labelled_image = fill_labelled_image(
        predicted_labels, coordinates, tile_size, img_width, img_height
    )
    return full_labelled_image


def draw_grid(data, img_width, img_height, tile_size):
    """Draw a white grid on an original image labelled version, depending on
    its tile splitting, *i.e.* draw vertical and horizontal lines each
    `tile_size` pixels.

    Parameters
    ----------
    data : numpy.array
        Labelled version of an image
    img_width : int
        Original image width, in pixel
    img_height : int
        Original image height, in pixel
    tile_size : int
        Size of the tiled images, in pixel

    Returns
    -------
    numpy.array
        Labelled version of an image, with an explicit white grid that
    highlights the tile splitting
    """
    gridded_data = data.copy()
    for i in range(tile_size, img_width, tile_size):
        gridded_data[:, i] = np.full([img_height, 3], 255, dtype=np.uint8)
    for i in range(tile_size, img_height, tile_size):
        gridded_data[i] = np.full([img_width, 3], 255, dtype=np.uint8)
    return gridded_data


def get_image_features(datapath, dataset, filename):
    """Retrieve geotiff image features with GDAL

    Use the `GetGeoTransform` method, that provides the following values:
      + East/West location of Upper Left corner
      + East/West pixel resolution
      + 0.0
      + North/South location of Upper Left corner
      + 0.0
      + North/South pixel resolution

    A GDAL dataset is opened during the function execution. The corresponding
    variable is set to None at the end of the function so as to free memory.

    See GDAL documentation (https://www.gdal.org/gdal_tutorial.html)

    Parameters
    ----------
    datapath : str
    dataset : str
    filename : str
        Name of the image file from which coordinates are extracted

    Returns
    -------
    dict
        Bounding box of the image (west, south, east, north coordinates), srid,
    and size (in pixels)

    """
    input_folder = utils.prepare_input_folder(datapath, dataset)
    filepath = os.path.join(
        input_folder, "testing", "images", filename + ".tif"
    )
    ds = gdal.Open(filepath)
    width = ds.RasterXSize
    height = ds.RasterYSize
    gt = ds.GetGeoTransform()
    minx = gt[0]
    miny = gt[3] + height * gt[5]
    maxx = gt[0] + width * gt[1]
    maxy = gt[3]
    srid = int(ds.GetProjection().split('"')[-2])
    ds = None  # Free memory used by the GDAL Dataset
    return {
        "west": minx,
        "south": miny,
        "east": maxx,
        "north": maxy,
        "srid": srid,
        "width": width,
        "height": height,
    }


def main(args):

    logger.info("Postprocess %s...", args.image_basename)
    features = get_image_features(
        args.datapath, args.dataset, args.image_basename
    )

    img_width, img_height = features["width"], features["height"]
    logger.info("Raw image size: %s, %s" % (img_width, img_height))

    prepro_folder = utils.prepare_preprocessed_folder(args.datapath, args.dataset, args.image_size)
    image_paths = get_image_paths(prepro_folder["testing"], args.image_basename)
    logger.info("The image will be splitted into %s tiles" % len(image_paths))
    images = extract_images(image_paths)
    coordinates = extract_coordinates_from_filenames(image_paths)
    labels = get_labels(args.datapath, args.dataset, args.image_size)

    output_folder = utils.prepare_output_folder(
        args.datapath, args.dataset, args.image_size, "semseg"
    )
    model = get_trained_model(
        output_folder["best-model"], args.image_size, len(labels)
    )

    logger.info("Predict labels for input images...")
    data = build_full_labelled_image(
        images,
        coordinates,
        model,
        args.image_size,
        img_width,
        img_height,
        args.batch_size,
    )
    logger.info(
        "Labelled image dimension: %s, %s" % (data.shape[0], data.shape[1])
    )
    colored_data = assign_label_colors(data, labels)
    if args.draw_grid:
        colored_data = draw_grid(
            colored_data, img_width, img_height, args.image_size
        )
    predicted_label_file = os.path.join(
        output_folder["labels"],
        args.image_basename + "_" + str(args.image_size) + ".png",
    )
    Image.fromarray(colored_data).save(predicted_label_file)

    vectorized_labels, vectorized_data = geometries.vectorize_mask(
        data, colored_data, labels
    )
    gdf = gpd.GeoDataFrame(
        {"labels": vectorized_labels, "geometry": vectorized_data}
    )
    predicted_geom_file = os.path.join(
        output_folder["geometries"],
        args.image_basename + "_" + str(args.image_size) + ".geojson",
    )
    if os.path.isfile(predicted_geom_file):
        os.remove(predicted_geom_file)
    gdf.to_file(predicted_geom_file, driver="GeoJSON")

    rasterized_data = geometries.rasterize_polygons(
        vectorized_data, vectorized_labels, img_height, img_width
    )
    colored_raster_data = assign_label_colors(rasterized_data, labels)
    if args.draw_grid:
        colored_raster_data = draw_grid(
            colored_raster_data, img_width, img_height, args.image_size
        )
    predicted_raster_file = os.path.join(
        output_folder["rasters"],
        args.image_basename + "_" + str(args.image_size) + ".png",
    )
    Image.fromarray(colored_raster_data).save(predicted_raster_file)



"""Script to aggregate a few Mapillary labels
"""

import os
import json

import pandas as pd
import seaborn as sns

from deeposlandia import utils


def set_label_color(nb_colors):
    """Set a color for each aggregated label with seaborn palettes

    Parameters
    ----------
    nb_colors : int
        Number of label to display
    """
    palette = sns.hls_palette(nb_colors, 0.01, 0.6, 0.75)
    return ([int(255 * item) for item in color] for color in palette)


def config_as_dataframe(config):
    """JSON labels data into a DataFrame.

    Add some metadata. Group some labels (in order to have less classes)

    Parameters
    ----------
    config : dict
        Mapillary glossary

    Returns
    -------
    pandas dataframe
        Mapillary labels
    """
    df = pd.DataFrame(config["labels"])
    df["id"] = range(df.shape[0])
    df["family"] = df["name"].str.split("--").apply(lambda x: x[0])
    df["label"] = df["name"].str.split("--").apply(lambda x: x[-1])
    df["new_label"] = df["name"].str.split("--").apply(lambda x: x[-2])
    return df


def aggregate_config(config, df):
    """Aggregate the labels from the original configuration

    Parameters
    ----------
    config : dict
    df : DataFrame

    Returns
    -------
    dict to serialize
    """
    assert len(config["labels"]) == df.shape[0], "should have the same size"
    result = {"folder_structure": config["folder_structure"]}
    result["labels"] = []
    nb_labels = df.new_label.nunique()
    intgen = iter(range(nb_labels))
    palette = set_label_color(nb_labels)
    for key, group in df.groupby("new_label"):
        d = {
            "id": next(intgen),
            "family": group.family.iloc[0],
            "name": group.new_label.iloc[0],
            "contains": group.label.str.cat(sep="..."),
            "contains_id": list(group.id),
            "color": next(palette),
            "evaluate": any(group.evaluate),
        }
        result["labels"].append(d)
    return result


def main(datadir):
    """Generate a new config.json file with aggregated labels.

    Parameters
    ----------
    datadir : str

    Returns
    -------
    dict
    """
    config = utils.read_config(os.path.join(datadir, "config.json"))
    df = config_as_dataframe(config)
    agg_config = aggregate_config(config, df)
    return agg_config


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=("Aggregate labels on a training image")
    )
    parser.add_argument(
        "-dp",
        "--datapath",
        required=False,
        default="data",
        nargs="?",
        help="Relative path towards data directory",
    )
    parser.add_argument(
        "-s",
        "--save",
        required=False,
        default="config_aggregate.json",
        help="Name of the output JSON file.",
    )
    args = parser.parse_args()
    label_aggregated = main(args.datapath)
    with open(os.path.join(args.datapath, args.save), "w") as fobj:
        print(
            "write the file '{}'".format(
                os.path.join(args.datapath, args.save)
            )
        )
        json.dump(label_aggregated, fobj)
